# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from concurrent.futures import ThreadPoolExecutor
import logging
import threading
import warnings
import time
from typing import (
    Any,
    Union,
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Callable,
    cast,
    overload,
)
from typing_extensions import Literal

from ._client_base import ClientBase
from ._producer import EventHubProducer
from ._constants import ALL_PARTITIONS, TransportType
from ._tracing import TraceAttributes
from ._common import EventDataBatch, EventData
from ._buffered_producer import BufferedProducerDispatcher
from ._utils import set_event_partition_key
from .amqp import AmqpAnnotatedMessage
from .exceptions import ConnectError, EventHubError

if TYPE_CHECKING:
    from ssl import SSLContext
    from ._client_base import CredentialTypes

SendEventTypes = List[Union[EventData, AmqpAnnotatedMessage]]

_LOGGER = logging.getLogger(__name__)


class EventHubProducerClient(ClientBase):  # pylint: disable=client-accepts-api-version-keyword
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=client-method-missing-tracing-decorator
    """The EventHubProducerClient class defines a high level interface for
    sending events to the Azure Event Hubs service.

    :param str fully_qualified_namespace: The fully qualified host name for the Event Hubs namespace.
     This is likely to be similar to <yournamespace>.servicebus.windows.net
    :param str eventhub_name: The path of the specific Event Hub to connect the client to.
    :param credential: The credential object used for authentication which
     implements a particular interface for getting tokens. It accepts
     :class:`EventHubSharedKeyCredential<azure.eventhub.EventHubSharedKeyCredential>`, or credential objects generated
     by the azure-identity library and objects that implement the `get_token(self, *scopes)` method.
    :type credential: ~azure.core.credentials.TokenCredential or ~azure.core.credentials.AzureSasCredential
     or ~azure.core.credentials.AzureNamedKeyCredential
    :keyword bool buffered_mode: If True, the producer client will collect events in a buffer, efficiently batch,
     then publish. Default is False.
    :keyword buffer_concurrency: The ThreadPoolExecutor to be used for publishing events
     or the number of workers for the ThreadPoolExecutor. Default is None and a ThreadPoolExecutor
     with the default number of workers will be created per
     https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor
    :paramtype buffer_concurrency: ThreadPoolExecutor or int or None
    :keyword on_success: The callback to be called once a batch has been successfully published.
     The callback takes two parameters:
     - `events`: The list of events that have been successfully published
     - `partition_id`: The partition id that the events in the list have been published to.

     The callback function should be defined like: `on_success(events, partition_id)`.
     It is required when `buffered_mode` is True while optional if `buffered_mode` is False.
    :paramtype on_success: Optional[Callable[[SendEventTypes, Optional[str]], None]]
    :keyword on_error: The callback to be called once a batch has failed to be published.
     The callback function should be defined like: `on_error(events, partition_id, error)`, where:
     - `events`: The list of events that failed to be published,
     - `partition_id`: The partition id that the events in the list have been tried to be published to and
     - `error`: The exception related to the sending failure.

     If `buffered_mode` is False, `on_error` callback is optional and errors will be handled as follows:
     - If an `on_error` callback is passed during the producer client instantiation,
     then error information will be passed to the `on_error` callback, which will then be called.
     - If an `on_error` callback is not passed in during client instantiation,
     then the error will be raised by default.

     If `buffered_mode` is True, `on_error` callback is required and errors will be handled as follows:
     - If events fail to enqueue within the given timeout, then an error will be directly raised.
     - If events fail to send after enqueuing successfully, the `on_error` callback will be called.

    :paramtype on_error: Optional[Callable[[SendEventTypes, Optional[str], Exception], None]]
    :keyword int max_buffer_length: Buffered mode only.
     The total number of events per partition that can be buffered before a flush will be triggered.
     The default value is 1500 in buffered mode.
    :keyword Optional[float] max_wait_time: Buffered mode only.
     The amount of time to wait for a batch to be built with events in the buffer before publishing.
     The default value is 1 in buffered mode.
    :keyword bool logging_enable: Whether to output network trace logs to the logger. Default is `False`.
    :keyword float auth_timeout: The time in seconds to wait for a token to be authorized by the service.
     The default value is 60 seconds. If set to 0, no timeout will be enforced from the client.
    :keyword str user_agent: If specified, this will be added in front of the user agent string.
    :keyword retry_total: The total number of attempts to redo a failed operation when an error occurs. Default
     value is 3.
    :paramtype retry_total: int
    :keyword float retry_backoff_factor: A backoff factor to apply between attempts after the second try
     (most errors are resolved immediately by a second try without a delay).
     In fixed mode, retry policy will always sleep for {backoff factor}.
     In 'exponential' mode, retry policy will sleep for: `{backoff factor} * (2 ** ({number of total retries} - 1))`
     seconds. If the backoff_factor is 0.1, then the retry will sleep
     for [0.0s, 0.2s, 0.4s, ...] between retries. The default value is 0.8.
    :keyword float retry_backoff_max: The maximum back off time. Default value is 120 seconds (2 minutes).
    :keyword retry_mode: The delay behavior between retry attempts. Supported values are 'fixed' or 'exponential',
     where default is 'exponential'.
    :paramtype retry_mode: str
    :keyword float idle_timeout: Timeout, in seconds, after which this client will close the underlying connection
     if there is no activity. By default the value is None, meaning that the client will not shutdown due to inactivity
     unless initiated by the service.
    :keyword transport_type: The type of transport protocol that will be used for communicating with
     the Event Hubs service. Default is `TransportType.Amqp` in which case port 5671 is used.
     If the port 5671 is unavailable/blocked in the network environment, `TransportType.AmqpOverWebsocket` could
     be used instead which uses port 443 for communication.
    :paramtype transport_type: ~azure.eventhub.TransportType
    :keyword Dict http_proxy: HTTP proxy settings. This must be a dictionary with the following
     keys: `'proxy_hostname'` (str value) and `'proxy_port'` (int value).
     Additionally the following keys may also be present: `'username', 'password'`.
    :keyword custom_endpoint_address: The custom endpoint address to use for establishing a connection to
     the Event Hubs service, allowing network requests to be routed through any application gateways or
     other paths needed for the host environment. Default is None.
     Unless specified otherwise, default transport type is TransportType.AmqpOverWebsockets.
     The format would be like "sb://<custom_endpoint_hostname>:<custom_endpoint_port>".
     If port is not specified in the `custom_endpoint_address`, by default port 443 will be used.
    :paramtype custom_endpoint_address: Optional[str]
    :keyword connection_verify: Path to the custom CA_BUNDLE file of the SSL certificate which is used to
     authenticate the identity of the connection endpoint.
     Default is None in which case `certifi.where()` will be used.
    :paramtype connection_verify: Optional[str]
    :keyword ssl_context: The SSLContext object to use in the underlying Pure Python AMQP transport. If specified,
     connection_verify will be ignored.
    :paramtype ssl_context: ssl.SSLContext or None
    :keyword uamqp_transport: Whether to use the `uamqp` library as the underlying transport. The default value is
     False and the Pure Python AMQP library will be used as the underlying transport.
    :paramtype uamqp_transport: bool
    :keyword float socket_timeout: The time in seconds that the underlying socket on the connection should
     wait when sending and receiving data before timing out. The default value is 0.2 for TransportType.Amqp
     and 1 for TransportType.AmqpOverWebsocket. If EventHubsConnectionError errors are occurring due to write
     timing out, a larger than default value may need to be passed in. This is for advanced usage scenarios
     and ordinarily the default value should be sufficient.

    .. admonition:: Example:

        .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
            :start-after: [START create_eventhub_producer_client_sync]
            :end-before: [END create_eventhub_producer_client_sync]
            :language: python
            :dedent: 4
            :caption: Create a new instance of the EventHubProducerClient.

    """

    @overload
    def __init__(
        self,
        fully_qualified_namespace: str,
        eventhub_name: str,
        credential: "CredentialTypes",
        *,
        buffered_mode: Literal[False] = False,
        **kwargs: Any,
    ) -> None: ...

    @overload
    def __init__(
        self,
        fully_qualified_namespace: str,
        eventhub_name: str,
        credential: "CredentialTypes",
        *,
        buffered_mode: Literal[True],
        buffer_concurrency: Optional[Union[ThreadPoolExecutor, int]] = None,
        on_error: Callable[[SendEventTypes, Optional[str], Exception], None],
        on_success: Callable[[SendEventTypes, Optional[str]], None],
        max_buffer_length: int = 1500,
        max_wait_time: float = 1,
        **kwargs: Any,
    ) -> None: ...

    def __init__(
        self,
        fully_qualified_namespace: str,
        eventhub_name: str,
        credential: "CredentialTypes",
        *,
        buffered_mode: bool = False,
        on_error: Optional[Callable[[SendEventTypes, Optional[str], Exception], None]] = None,
        on_success: Optional[Callable[[SendEventTypes, Optional[str]], None]] = None,
        max_buffer_length: Optional[int] = None,
        max_wait_time: Optional[float] = None,
        **kwargs: Any,
    ) -> None:
        super(EventHubProducerClient, self).__init__(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=credential,
            network_tracing=kwargs.get("logging_enable"),
            **kwargs,
        )
        # Deprecation of uamqp transport
        if kwargs.get("uamqp_transport"):
            warnings.warn(
                "uAMQP legacy support will be removed in the 5.16.0 minor release. "
                "Please remove the use of `uamqp_transport` keyword argument from the client in order "
                "to use the pure Python AMQP transport. "
                "If you rely on this, please comment on [this issue]"
                "(https://github.com/Azure/azure-sdk-for-python/issues/40347) ",
                DeprecationWarning, stacklevel=2
            )
        self._auth_uri = f"sb://{self._address.hostname}{self._address.path}"
        self._keep_alive = kwargs.get("keep_alive", None)

        self._producers: Dict[str, Optional[EventHubProducer]] = {ALL_PARTITIONS: self._create_producer()}
        self._max_message_size_on_link = 0
        self._partition_ids: Optional[List[str]] = None
        self._lock = threading.Lock()
        self._buffered_mode = buffered_mode
        self._on_success = on_success
        self._on_error = on_error
        self._buffered_producer_dispatcher = None
        self._max_wait_time = max_wait_time
        self._max_buffer_length = max_buffer_length
        self._executor = kwargs.get("buffer_concurrency")

        if self._buffered_mode:
            setattr(self, "send_batch", self._buffered_send_batch)
            setattr(self, "send_event", self._buffered_send_event)
            if not self._on_error:
                raise TypeError(
                    "EventHubProducerClient in buffered mode missing 1 required keyword argument: 'on_error'"
                )
            if not self._on_success:
                raise TypeError(
                    "EventHubProducerClient in buffered mode missing 1 required keyword argument: 'on_success'"
                )
            if self._max_wait_time is None:
                self._max_wait_time = 1
            if self._max_wait_time <= 0:
                raise ValueError("'max_wait_time' must be a float greater than 0 in buffered mode")
            if self._max_buffer_length is None:
                self._max_buffer_length = 1500
            if self._max_buffer_length <= 0:
                raise ValueError("'max_buffer_length' must be an integer greater than 0 in buffered mode")
            if isinstance(self._executor, int) and self._executor <= 0:
                raise ValueError("'buffer_concurrency' must be an integer greater than 0 in buffered mode")

    def __enter__(self) -> "EventHubProducerClient":
        return self

    def __exit__(self, *args: Any) -> None:
        self.close()

    def _buffered_send(self, events, **kwargs):
        try:
            self._buffered_producer_dispatcher.enqueue_events(events, **kwargs)
        except AttributeError:
            self._get_partitions()
            self._get_max_message_size()
            self._buffered_producer_dispatcher = BufferedProducerDispatcher(
                self._partition_ids,
                self._on_success,
                self._on_error,
                self._create_producer,
                self.eventhub_name,
                self._max_message_size_on_link,
                max_wait_time=self._max_wait_time,
                max_buffer_length=self._max_buffer_length,
                executor=self._executor,
                amqp_transport=self._amqp_transport,
            )
            self._buffered_producer_dispatcher.enqueue_events(events, **kwargs)

    def _batch_preparer(
        self,
        event_data_batch: Union[EventDataBatch, SendEventTypes],
        *,
        partition_id: Optional[str] = None,
        partition_key: Optional[str] = None,
    ):
        if isinstance(event_data_batch, EventDataBatch):
            if partition_id or partition_key:
                raise TypeError(
                    "partition_id and partition_key should be None when sending an EventDataBatch "
                    "because type EventDataBatch itself may have partition_id or partition_key"
                )
            to_send_batch = event_data_batch
        else:
            to_send_batch = self.create_batch(partition_id=partition_id, partition_key=partition_key)
            to_send_batch._load_events(event_data_batch)  # pylint:disable=protected-access

        return (
            to_send_batch,
            to_send_batch._partition_id,  # pylint:disable=protected-access
            partition_key,
        )

    def _buffered_send_batch(  # pylint: disable=unused-argument
        self,
        event_data_batch: Union[EventDataBatch, SendEventTypes],
        *,
        timeout: Optional[float] = None,
        partition_key: Optional[str] = None,
        partition_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        batch, pid, pkey = self._batch_preparer(
            event_data_batch, partition_id=partition_id, partition_key=partition_key
        )

        if len(batch) == 0:
            return

        timeout_time = time.time() + timeout if timeout else None

        self._buffered_send(
            event_data_batch,
            partition_id=pid,
            partition_key=pkey,
            timeout_time=timeout_time,
        )

    def _buffered_send_event(
        self,  # pylint: disable=unused-argument
        event: Union[EventData, AmqpAnnotatedMessage],
        *,
        timeout: Optional[float] = None,
        partition_id: Optional[str] = None,
        partition_key: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        set_event_partition_key(event, partition_key, self._amqp_transport)
        timeout_time = time.time() + timeout if timeout else None
        self._buffered_send(
            event,
            partition_id=partition_id,
            partition_key=partition_key,
            timeout_time=timeout_time,
        )

    def _get_partitions(self) -> None:
        if not self._partition_ids:
            _LOGGER.debug("Populating partition IDs so producers can be started.")
            self._partition_ids = self.get_partition_ids()  # type: ignore
            for p_id in cast(List[str], self._partition_ids):
                self._producers[p_id] = None

    def _get_max_message_size(self) -> None:
        # pylint: disable=protected-access
        with self._lock:
            if not self._max_message_size_on_link:
                cast(EventHubProducer, self._producers[ALL_PARTITIONS])._open_with_retry()
                self._max_message_size_on_link = (
                    self._amqp_transport.get_remote_max_message_size(
                        self._producers[ALL_PARTITIONS]._handler  # type: ignore
                    )
                    or self._amqp_transport.MAX_MESSAGE_LENGTH_BYTES
                )

    def _start_producer(self, partition_id: str, send_timeout: Optional[float] = None) -> None:
        with self._lock:
            self._get_partitions()
            if partition_id not in cast(List[str], self._partition_ids) and partition_id != ALL_PARTITIONS:
                raise ConnectError("Invalid partition {} for the event hub {}".format(partition_id, self.eventhub_name))

            if not self._producers[partition_id] or cast(EventHubProducer, self._producers[partition_id]).closed:
                self._producers[partition_id] = self._create_producer(
                    partition_id=(None if partition_id == ALL_PARTITIONS else partition_id),
                    send_timeout=send_timeout,
                )

    def _create_producer(
        self, partition_id: Optional[str] = None, send_timeout: Optional[float] = None
    ) -> EventHubProducer:
        target = "amqps://{}{}".format(self._address.hostname, self._address.path)
        send_timeout = self._config.send_timeout if send_timeout is None else send_timeout

        handler = EventHubProducer(
            self,
            target,
            partition=partition_id,
            send_timeout=send_timeout,
            idle_timeout=self._idle_timeout,
            amqp_transport=self._amqp_transport,
            keep_alive=self._keep_alive,
        )
        return handler

    @classmethod
    @overload
    def from_connection_string(
        cls, conn_str: str, *, eventhub_name: Optional[str] = None, buffered_mode: Literal[False] = False, **kwargs: Any
    ) -> "EventHubProducerClient": ...

    @classmethod
    @overload
    def from_connection_string(
        cls,
        conn_str: str,
        *,
        eventhub_name: Optional[str] = None,
        buffered_mode: Literal[True],
        buffer_concurrency: Optional[Union[ThreadPoolExecutor, int]] = None,
        on_error: Callable[[SendEventTypes, Optional[str], Exception], None],
        on_success: Callable[[SendEventTypes, Optional[str]], None],
        max_buffer_length: int = 1500,
        max_wait_time: float = 1,
        **kwargs: Any,
    ) -> "EventHubProducerClient": ...

    @classmethod
    def from_connection_string(
        cls,
        conn_str: str,
        *,
        eventhub_name: Optional[str] = None,
        buffered_mode: bool = False,
        buffer_concurrency: Optional[Union[ThreadPoolExecutor, int]] = None,
        on_success: Optional[Callable[[SendEventTypes, Optional[str]], None]] = None,
        on_error: Optional[Callable[[SendEventTypes, Optional[str], Exception], None]] = None,
        max_buffer_length: Optional[int] = None,
        max_wait_time: Optional[float] = None,
        logging_enable: bool = False,
        http_proxy: Optional[Dict[str, Union[str, int]]] = None,
        auth_timeout: Optional[float] = 60,
        user_agent: Optional[str] = None,
        retry_total: Optional[int] = 3,
        retry_backoff_factor: Optional[float] = 0.8,
        retry_backoff_max: Optional[float] = 120,
        retry_mode: Literal["exponential", "fixed"] = "exponential",
        idle_timeout: Optional[float] = None,
        transport_type: Optional["TransportType"] = TransportType.Amqp,
        custom_endpoint_address: Optional[str] = None,
        connection_verify: Optional[str] = None,
        ssl_context: Optional["SSLContext"] = None,
        uamqp_transport: bool = False,
        **kwargs: Any,
    ) -> "EventHubProducerClient":
        """Create an EventHubProducerClient from a connection string.

        :param str conn_str: The connection string of an Event Hub.
        :keyword str eventhub_name: The path of the specific Event Hub to connect the client to.
        :keyword bool buffered_mode: If True, the producer client will collect events in a buffer, efficiently batch,
         then publish. Default is False.
        :keyword buffer_concurrency: The ThreadPoolExecutor to be used for publishing events
         or the number of workers for the ThreadPoolExecutor. Default is None and a ThreadPoolExecutor
         with the default number of workers will be created per
         https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor
        :paramtype buffer_concurrency: ThreadPoolExecutor or int or None
        :keyword on_success: The callback to be called once a batch has been successfully published.
         The callback takes two parameters:
         - `events`: The list of events that have been successfully published
         - `partition_id`: The partition id that the events in the list have been published to.

         The callback function should be defined like: `on_success(events, partition_id)`.
         Required when `buffered_mode` is True while optional if `buffered_mode` is False.
        :paramtype on_success: Optional[Callable[[SendEventTypes, Optional[str]], None]]
        :keyword on_error: The callback to be called once a batch has failed to be published.
         Required when in `buffered_mode` is True while optional if `buffered_mode` is False.
         The callback function should be defined like: `on_error(events, partition_id, error)`, where:
         - `events`: The list of events that failed to be published,
         - `partition_id`: The partition id that the events in the list have been tried to be published to and
         - `error`: The exception related to the sending failure.

         If `buffered_mode` is False, `on_error` callback is optional and errors will be handled as follows:
         - If an `on_error` callback is passed during the producer client instantiation,
         then error information will be passed to the `on_error` callback, which will then be called.
         - If an `on_error` callback is not passed in during client instantiation,
         then the error will be raised by default.

         If `buffered_mode` is True, `on_error` callback is required and errors will be handled as follows:
         - If events fail to enqueue within the given timeout, then an error will be directly raised.
         - If events fail to send after enqueuing successfully, the `on_error` callback will be called.

        :paramtype on_error: Optional[Callable[[SendEventTypes, Optional[str], Exception], None]]
        :keyword int max_buffer_length: Buffered mode only.
         The total number of events per partition that can be buffered before a flush will be triggered.
         The default value is 1500 in buffered mode.
        :keyword Optional[float] max_wait_time: Buffered mode only.
         The amount of time to wait for a batch to be built with events in the buffer before publishing.
         The default value is 1 in buffered mode.
        :keyword bool logging_enable: Whether to output network trace logs to the logger. Default is `False`.
        :keyword dict http_proxy: HTTP proxy settings. This must be a dictionary with the following
         keys: `'proxy_hostname'` (str value) and `'proxy_port'` (int value).
         Additionally the following keys may also be present: `'username', 'password'`.
        :keyword float auth_timeout: The time in seconds to wait for a token to be authorized by the service.
         The default value is 60 seconds. If set to 0, no timeout will be enforced from the client.
        :keyword str user_agent: If specified, this will be added in front of the user agent string.
        :keyword retry_total: The total number of attempts to redo a failed operation when an error occurs.
         Default value is 3.
        :paramtype retry_total: int
        :keyword float retry_backoff_factor: A backoff factor to apply between attempts after the second try
         (most errors are resolved immediately by a second try without a delay).
         In fixed mode, retry policy will always sleep for {backoff factor}.
         In 'exponential' mode, retry policy will sleep for: `{backoff factor} * (2 ** ({number of total retries} - 1))`
         seconds. If the backoff_factor is 0.1, then the retry will sleep
         for [0.0s, 0.2s, 0.4s, ...] between retries. The default value is 0.8.
        :keyword float retry_backoff_max: The maximum back off time. Default value is 120 seconds (2 minutes).
        :keyword retry_mode: The delay behavior between retry attempts. Supported values are 'fixed' or 'exponential',
         where default is 'exponential'.
        :paramtype retry_mode: str
        :keyword float idle_timeout: Timeout, in seconds, after which this client will close the underlying connection
         if there is no activity. By default the value is None, meaning that the client will not shutdown due to
         inactivity unless initiated by the service.
        :keyword transport_type: The type of transport protocol that will be used for communicating with
         the Event Hubs service. Default is `TransportType.Amqp` in which case port 5671 is used.
         If the port 5671 is unavailable/blocked in the network environment, `TransportType.AmqpOverWebsocket` could
         be used instead which uses port 443 for communication.
        :paramtype transport_type: ~azure.eventhub.TransportType
        :keyword Dict http_proxy: HTTP proxy settings. This must be a dictionary with the following
         keys: `'proxy_hostname'` (str value) and `'proxy_port'` (int value).
         Additionally the following keys may also be present: `'username', 'password'`.
        :keyword custom_endpoint_address: The custom endpoint address to use for establishing a connection to
         the Event Hubs service, allowing network requests to be routed through any application gateways or
         other paths needed for the host environment. Default is None.
         Unless specified otherwise, default transport type is TransportType.AmqpOverWebsockets.
         The format would be like "sb://<custom_endpoint_hostname>:<custom_endpoint_port>".
         If port is not specified in the `custom_endpoint_address`, by default port 443 will be used.
        :paramtype custom_endpoint_address: Optional[str]
        :keyword connection_verify: Path to the custom CA_BUNDLE file of the SSL certificate which is used to
         authenticate the identity of the connection endpoint.
         Default is None in which case `certifi.where()` will be used.
        :paramtype connection_verify: Optional[str]
        :keyword ssl_context: The SSLContext object to use in the underlying Pure Python AMQP transport. If specified,
         connection_verify will be ignored.
        :paramtype ssl_context: ssl.SSLContext or None
        :keyword uamqp_transport: Whether to use the `uamqp` library as the underlying transport. The default value is
         False and the Pure Python AMQP library will be used as the underlying transport.
        :paramtype uamqp_transport: bool
        :returns: An instance of the EventHubProducerClient.
        :rtype: ~azure.eventhub.EventHubProducerClient

        .. admonition:: Example:

            .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
                :start-after: [START create_eventhub_producer_client_from_conn_str_sync]
                :end-before: [END create_eventhub_producer_client_from_conn_str_sync]
                :language: python
                :dedent: 4
                :caption: Create a new instance of the EventHubProducerClient from connection string.
        """
        constructor_args = cls._from_connection_string(
            conn_str,
            eventhub_name=eventhub_name,
            buffered_mode=buffered_mode,
            buffer_concurrency=buffer_concurrency,
            on_success=on_success,
            on_error=on_error,
            max_buffer_length=max_buffer_length,
            max_wait_time=max_wait_time,
            logging_enable=logging_enable,
            http_proxy=http_proxy,
            auth_timeout=auth_timeout,
            user_agent=user_agent,
            retry_total=retry_total,
            retry_backoff_factor=retry_backoff_factor,
            retry_backoff_max=retry_backoff_max,
            retry_mode=retry_mode,
            idle_timeout=idle_timeout,
            transport_type=transport_type,
            custom_endpoint_address=custom_endpoint_address,
            connection_verify=connection_verify,
            ssl_context=ssl_context,
            uamqp_transport=uamqp_transport,
            **kwargs,
        )
        return cls(**constructor_args)

    def send_event(
        self,  # pylint: disable=unused-argument
        event_data: Union[EventData, AmqpAnnotatedMessage],
        *,
        timeout: Optional[float] = None,
        partition_id: Optional[str] = None,
        partition_key: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Sends an event data.
        By default, the method will block until acknowledgement is received or operation times out.
        If the `EventHubProducerClient` is configured to run in buffered mode, the method will try enqueuing
        the events into buffer within the given time if specified and return.
        The producer will do automatic sending in the background in buffered mode.

         If `buffered_mode` is False, `on_error` callback is optional and errors will be handled as follows:
            - If an `on_error` callback is passed during the producer client instantiation,
                then error information will be passed to the `on_error` callback, which will then be called.
            - If an `on_error` callback is not passed in during client instantiation,
                then the error will be raised by default.

         If `buffered_mode` is True, `on_error` callback is required and errors will be handled as follows:
            - If events fail to enqueue within the given timeout, then an error will be directly raised.
            - If events fail to send after enqueuing successfully, the `on_error` callback will be called.

        :param event_data: The `EventData` object to be sent.
        :type event_data: Union[~azure.eventhub.EventData, ~azure.eventhub.amqp.AmqpAnnotatedMessage]
        :keyword float timeout: The maximum wait time to send the event data in non-buffered mode or the
         maximum wait time to enqueue the event data into the buffer in buffered mode.
         In non-buffered mode, the default wait time specified when the producer
         was created will be used. In buffered mode, the default wait time is None indicating that the event will be
         scheduled to send immediately.
        :keyword str partition_id: The specific partition ID to send to. Default is None, in which case the service
         will assign to all partitions using round-robin.
         A `TypeError` will be raised if partition_id is specified and event_data_batch is an `EventDataBatch` because
         `EventDataBatch` itself has partition_id.
        :keyword str partition_key: With the given partition_key, event data will be sent to
         a particular partition of the Event Hub decided by the service.
         A `TypeError` will be raised if partition_key is specified and event_data_batch is an `EventDataBatch` because
         `EventDataBatch` itself has partition_key.
         If both partition_id and partition_key are provided, the partition_id will take precedence.
         **WARNING: Setting partition_key of non-string value on the events to be sent is discouraged
         as the partition_key will be ignored by the Event Hub service and events will be assigned
         to all partitions using round-robin. Furthermore, there are SDKs for consuming events which expect
         partition_key to only be string type, they might fail to parse the non-string value.**
        :rtype: None
        :raises: :class:`AuthenticationError<azure.eventhub.exceptions.AuthenticationError>`
         :class:`ConnectError<azure.eventhub.exceptions.ConnectError>`
         :class:`ConnectionLostError<azure.eventhub.exceptions.ConnectionLostError>`
         :class:`EventDataError<azure.eventhub.exceptions.EventDataError>`
         :class:`EventDataSendError<azure.eventhub.exceptions.EventDataSendError>`
         :class:`EventHubError<azure.eventhub.exceptions.EventHubError>`
        :raises OperationTimeoutError: If the value specified by the timeout parameter elapses before the event can be
         sent in non-buffered mode or the events can be enqueued into the buffered in buffered mode.
        """
        input_pid = partition_id
        pid: str = input_pid or ALL_PARTITIONS
        send_timeout = timeout
        try:
            try:
                cast(EventHubProducer, self._producers[pid]).send(
                    event_data, partition_key=partition_key, timeout=send_timeout
                )
            except (KeyError, AttributeError, EventHubError):
                self._start_producer(pid, send_timeout)
                cast(EventHubProducer, self._producers[pid]).send(
                    event_data, partition_key=partition_key, timeout=send_timeout
                )
            if self._on_success:
                self._on_success([event_data], input_pid)
        except Exception as exc:  # pylint: disable=broad-except
            if self._on_error:
                self._on_error([event_data], input_pid, exc)
            else:
                raise

    def send_batch(  # pylint: disable=unused-argument
        self,
        event_data_batch: Union[EventDataBatch, SendEventTypes],
        *,
        timeout: Optional[float] = None,
        partition_key: Optional[str] = None,
        partition_id: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        # pylint: disable=protected-access
        """
        Sends a batch of event data.
        By default, the method will block until acknowledgement is received or operation times out.
        If the `EventHubProducerClient` is configured to run in buffered mode, the method will try enqueuing
        the events into buffer within the given time if specified and return.
        The producer will do automatic sending in the background in buffered mode.

        If `buffered_mode` is False, `on_error` callback is optional and errors will be handled as follows:
           - If an `on_error` callback is passed during the producer client instantiation,
               then error information will be passed to the `on_error` callback, which will then be called.
           - If an `on_error` callback is not passed in during client instantiation,
               then the error will be raised by default.

        If `buffered_mode` is True, `on_error` callback is required and errors will be handled as follows:
           - If events fail to enqueue within the given timeout, then an error will be directly raised.
           - If events fail to send after enqueuing successfully, the `on_error` callback will be called.

        In buffered mode, sending a batch will remain intact and sent as a single unit.
        The batch will not be rearranged. This may result in inefficiency of sending events.

        If you're sending a finite list of `EventData` or `AmqpAnnotatedMessage` and you know it's within the
        event hub frame size limit, you can send them with a `send_batch` call. Otherwise, use :meth:`create_batch`
        to create `EventDataBatch` and add either `EventData` or `AmqpAnnotatedMessage` into the batch one by one
        until the size limit, and then call this method to send out the batch.

        :param event_data_batch: The `EventDataBatch` object to be sent or a list of `EventData` to be sent in a batch.
         All `EventData` or `AmqpAnnotatedMessage` in the list or `EventDataBatch` will land on the same partition.
        :type event_data_batch: Union[~azure.eventhub.EventDataBatch, List[Union[~azure.eventhub.EventData,
            ~azure.eventhub.amqp.AmqpAnnotatedMessage]]
        :keyword float timeout: The maximum wait time to send the event data in non-buffered mode or the
         maximum wait time to enqueue the event data into the buffer in buffered mode.
         In non-buffered mode, the default wait time specified when the producer
         was created will be used. In buffered mode, the default wait time is None.
        :keyword str partition_id: The specific partition ID to send to. Default is None, in which case the service
         will assign to all partitions using round-robin.
         A `TypeError` will be raised if partition_id is specified and event_data_batch is an `EventDataBatch` because
         `EventDataBatch` itself has partition_id.
        :keyword str partition_key: With the given partition_key, event data will be sent to
         a particular partition of the Event Hub decided by the service.
         A `TypeError` will be raised if partition_key is specified and event_data_batch is an `EventDataBatch` because
         `EventDataBatch` itself has partition_key.
         If both partition_id and partition_key are provided, the partition_id will take precedence.
         **WARNING: Setting partition_key of non-string value on the events to be sent is discouraged
         as the partition_key will be ignored by the Event Hub service and events will be assigned
         to all partitions using round-robin. Furthermore, there are SDKs for consuming events which expect
         partition_key to only be string type, they might fail to parse the non-string value.**
        :rtype: None
        :raises: :class:`AuthenticationError<azure.eventhub.exceptions.AuthenticationError>`
         :class:`ConnectError<azure.eventhub.exceptions.ConnectError>`
         :class:`ConnectionLostError<azure.eventhub.exceptions.ConnectionLostError>`
         :class:`EventDataError<azure.eventhub.exceptions.EventDataError>`
         :class:`EventDataSendError<azure.eventhub.exceptions.EventDataSendError>`
         :class:`EventHubError<azure.eventhub.exceptions.EventHubError>`
         :class:`ValueError`
         :class:`TypeError`
        :raises OperationTimeoutError: If the value specified by the timeout parameter elapses before the event can be
         sent in non-buffered mode or the events can not be enqueued into the buffered in buffered mode.

        .. admonition:: Example:

            .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
                :start-after: [START eventhub_producer_client_send_sync]
                :end-before: [END eventhub_producer_client_send_sync]
                :language: python
                :dedent: 4
                :caption: Sends event data

        """
        batch, pid, pkey = self._batch_preparer(
            event_data_batch, partition_id=partition_id, partition_key=partition_key
        )

        if len(batch) == 0:
            return

        partition_id = pid or ALL_PARTITIONS
        send_timeout = timeout

        try:
            try:
                cast(EventHubProducer, self._producers[partition_id]).send(
                    batch, partition_key=pkey, timeout=send_timeout
                )
                if self._on_success:
                    self._on_success(batch._internal_events, pid)
            except (KeyError, AttributeError, EventHubError) as e:
                _LOGGER.debug(
                    "Producer for partition ID %s not available: %s. Rebuilding new producer.",
                    partition_id,
                    e,
                )
                self._start_producer(partition_id, send_timeout)
                cast(EventHubProducer, self._producers[partition_id]).send(
                    batch, partition_key=pkey, timeout=send_timeout
                )
                if self._on_success:
                    self._on_success(batch._internal_events, pid)
        except Exception as exc:  # pylint: disable=broad-except
            if self._on_error:
                self._on_error(batch._internal_events, pid, exc)
            else:
                raise

    def create_batch(  # pylint: disable=unused-argument
        self,
        *,
        partition_id: Optional[str] = None,
        partition_key: Optional[str] = None,
        max_size_in_bytes: Optional[int] = None,
        **kwargs: Any,
    ) -> EventDataBatch:
        """Create an EventDataBatch object with the max size of all content being constrained by max_size_in_bytes.

        The max_size_in_bytes should be no greater than the max allowed message size defined by the service.

        :keyword str partition_id: The specific partition ID to send to. Default is None, in which case the service
         will assign to all partitions using round-robin.
        :keyword str partition_key: With the given partition_key, event data will be sent to
         a particular partition of the Event Hub decided by the service.
         If both partition_id and partition_key are provided, the partition_id will take precedence.
         **WARNING: Setting partition_key of non-string value on the events to be sent is discouraged
         as the partition_key will be ignored by the Event Hub service and events will be assigned
         to all partitions using round-robin. Furthermore, there are SDKs for consuming events which expect
         partition_key to only be string type, they might fail to parse the non-string value.**
        :keyword int max_size_in_bytes: The maximum size of bytes data that an EventDataBatch object can hold. By
         default, the value is determined by your Event Hubs tier.
        :return: An EventDataBatch object instance

        .. admonition:: Example:

            .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
                :start-after: [START eventhub_producer_client_create_batch_sync]
                :end-before: [END eventhub_producer_client_create_batch_sync]
                :language: python
                :dedent: 4
                :caption: Create EventDataBatch object within limited size

        :rtype: ~azure.eventhub.EventDataBatch
        """
        if not self._max_message_size_on_link:
            self._get_max_message_size()

        if max_size_in_bytes and max_size_in_bytes > self._max_message_size_on_link:
            raise ValueError(
                "Max message size: {} is too large, acceptable max batch size is: {} bytes.".format(
                    max_size_in_bytes, self._max_message_size_on_link
                )
            )

        return EventDataBatch(
            max_size_in_bytes=(max_size_in_bytes or self._max_message_size_on_link),
            partition_id=partition_id,
            partition_key=partition_key,
            amqp_transport=self._amqp_transport,
            tracing_attributes={
                TraceAttributes.TRACE_NET_PEER_NAME_ATTRIBUTE: self._address.hostname if self._address else None,
                TraceAttributes.TRACE_MESSAGING_DESTINATION_ATTRIBUTE: self._address.path if self._address else None,
            },
        )

    def get_eventhub_properties(self) -> Dict[str, Any]:
        """Get properties of the Event Hub.

        Keys in the returned dictionary include:

            - `eventhub_name` (str)
            - `created_at` (UTC datetime.datetime)
            - `partition_ids` (list[str])

        :return: A dictionary containing eventhub properties.
        :rtype: dict[str, any]
        :raises: :class:`EventHubError<azure.eventhub.exceptions.EventHubError>`
        """
        return super(EventHubProducerClient, self)._get_eventhub_properties()

    def get_partition_ids(self) -> List[str]:
        """Get partition IDs of the Event Hub.

        :return: A list of partition IDs.
        :rtype: list[str]
        :raises: :class:`EventHubError<azure.eventhub.exceptions.EventHubError>`
        """
        return super(EventHubProducerClient, self)._get_partition_ids()

    def get_partition_properties(self, partition_id: str) -> Dict[str, Any]:
        """Get properties of the specified partition.

        Keys in the properties dictionary include:

            - `eventhub_name` (str)
            - `id` (str)
            - `beginning_sequence_number` (int)
            - `last_enqueued_sequence_number` (int)
            - `last_enqueued_offset` (str)
            - `last_enqueued_time_utc` (UTC datetime.datetime)
            - `is_empty` (bool)

        :param partition_id: The target partition ID.
        :type partition_id: str
        :return: A dictionary of partition properties.
        :rtype: dict[str, any]
        :raises: :class:`EventHubError<azure.eventhub.exceptions.EventHubError>`
        """
        return super(EventHubProducerClient, self)._get_partition_properties(partition_id)

    def flush(self, *, timeout: Optional[float] = None, **kwargs: Any) -> None:  # pylint:disable=unused-argument
        """
        Buffered mode only.
        Flush events in the buffer to be sent immediately if the client is working in buffered mode.

        :keyword Optional[float] timeout: Timeout to flush the buffered events, default is None which means no timeout.
        :rtype: None
        :raises EventDataSendError: If the producer fails to flush the buffer within the given timeout
         in buffered mode.
        """
        with self._lock:
            if self._buffered_mode and self._buffered_producer_dispatcher:
                timeout_time = time.time() + timeout if timeout else None
                self._buffered_producer_dispatcher.flush(timeout_time=timeout_time)

    def close(
        self, *, flush: bool = True, timeout: Optional[float] = None, **kwargs: Any  # pylint:disable=unused-argument
    ) -> None:
        """Close the Producer client underlying AMQP connection and links.

        :keyword bool flush: Buffered mode only. If set to True, events in the buffer will be sent
         immediately. Default is True.
        :keyword float or None timeout: Buffered mode only. Timeout to close the producer.
         Default is None which means no timeout.
        :rtype: None
        :raises EventHubError: If an error occurred when flushing the buffer if `flush` is set to True or closing the
         underlying AMQP connections in buffered mode.

        .. admonition:: Example:

            .. literalinclude:: ../samples/sync_samples/sample_code_eventhub.py
                :start-after: [START eventhub_producer_client_close_sync]
                :end-before: [END eventhub_producer_client_close_sync]
                :language: python
                :dedent: 4
                :caption: Close down the client.

        """
        with self._lock:
            if self._buffered_mode and self._buffered_producer_dispatcher:
                timeout_time = time.time() + timeout if timeout else None
                self._buffered_producer_dispatcher.close(flush=flush, timeout_time=timeout_time, raise_error=True)
                self._buffered_producer_dispatcher = None

            for pid, producer in self._producers.items():
                if producer:
                    producer.close()  # type: ignore
                self._producers[pid] = None
        super(EventHubProducerClient, self)._close()

    def get_buffered_event_count(self, partition_id: str) -> Optional[int]:
        """
        The number of events that are buffered and waiting to be published for a given partition.
         Returns None in non-buffered mode. **NOTE: The event buffer is processed in a background
         thread, therefore the number of events in the buffer reported by this API should be
         considered only an approximation and is only recommend for use in debugging. For a
         partition ID that has no events buffered, 0 will be returned regardless of whether that
         partition ID actually exists within the Event Hub.**


        :param str partition_id: The target partition ID.
        :rtype: int or None
        """
        if not self._buffered_mode:
            return None

        try:
            return cast(BufferedProducerDispatcher, self._buffered_producer_dispatcher).get_buffered_event_count(
                partition_id
            )
        except AttributeError:
            return 0

    @property
    def total_buffered_event_count(self) -> Optional[int]:
        """
        The total number of events that are currently buffered and waiting to be published,
         across all partitions. Returns None in non-buffered mode. **NOTE: The event buffer
         is processed in a background thread, therefore the number of events in the buffer
         reported by this API should be considered only an approximation and is only recommend
         for use in debugging.**

        :rtype: int or None
        """
        if not self._buffered_mode:
            return None

        try:
            return cast(BufferedProducerDispatcher, self._buffered_producer_dispatcher).total_buffered_event_count
        except AttributeError:
            return 0



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
Examples to show basic async use case of python azure-eventhub SDK, including:
    - Create EventHubProducerClient
    - Create EventHubConsumerClient
    - Create EventData
    - Create EventDataBatch
    - Send EventDataBatch
    - Receive EventData
    - Close EventHubProducerClient
    - Close EventHubConsumerClient
"""

import logging
import asyncio


def example_create_async_eventhub_producer_client():
    # [START create_eventhub_producer_client_from_conn_str_async]
    import os
    from azure.eventhub.aio import EventHubProducerClient
    from azure.identity.aio import DefaultAzureCredential

    fully_qualified_namespace = os.environ["EVENT_HUB_HOSTNAME"]
    eventhub_name = os.environ["EVENT_HUB_NAME"]
    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,  # EventHub name should be specified if it doesn't show up in connection string.
        credential=DefaultAzureCredential(),
    )
    # [END create_eventhub_producer_client_from_conn_str_async]

    # [START create_eventhub_producer_client_async]
    import os
    from azure.eventhub.aio import EventHubProducerClient, EventHubSharedKeyCredential

    fully_qualified_namespace = os.environ["EVENT_HUB_HOSTNAME"]
    eventhub_name = os.environ["EVENT_HUB_NAME"]
    shared_access_policy = os.environ["EVENT_HUB_SAS_POLICY"]
    shared_access_key = os.environ["EVENT_HUB_SAS_KEY"]

    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=EventHubSharedKeyCredential(shared_access_policy, shared_access_key),
    )
    # [END create_eventhub_producer_client_async]
    return producer


def example_create_async_eventhub_consumer_client():
    # [START create_eventhub_consumer_client_from_conn_str_async]
    import os
    from azure.eventhub.aio import EventHubConsumerClient
    from azure.identity.aio import DefaultAzureCredential

    fully_qualified_namespace = os.environ["EVENT_HUB_HOSTNAME"]
    eventhub_name = os.environ["EVENT_HUB_NAME"]
    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        credential=DefaultAzureCredential(),
        consumer_group="$Default",
        eventhub_name=eventhub_name,  # EventHub name should be specified if it doesn't show up in connection string.
    )
    # [END create_eventhub_consumer_client_from_conn_str_async]

    # [START create_eventhub_consumer_client_async]
    import os
    from azure.eventhub.aio import EventHubConsumerClient, EventHubSharedKeyCredential

    fully_qualified_namespace = os.environ["EVENT_HUB_HOSTNAME"]
    eventhub_name = os.environ["EVENT_HUB_NAME"]
    shared_access_policy = os.environ["EVENT_HUB_SAS_POLICY"]
    shared_access_key = os.environ["EVENT_HUB_SAS_KEY"]

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        consumer_group="$Default",
        eventhub_name=eventhub_name,
        credential=EventHubSharedKeyCredential(shared_access_policy, shared_access_key),
    )
    # [END create_eventhub_consumer_client_async]
    return consumer


async def example_eventhub_async_send_and_receive():
    producer = example_create_async_eventhub_producer_client()
    consumer = example_create_async_eventhub_consumer_client()
    try:
        # [START eventhub_producer_client_create_batch_async]
        from azure.eventhub import EventData

        event_data_batch = await producer.create_batch()
        while True:
            try:
                event_data_batch.add(EventData("Message inside EventBatchData"))
            except ValueError:
                # The EventDataBatch object reaches its max_size.
                # You can send the full EventDataBatch object and create a new one here.
                break
        # [END eventhub_producer_client_create_batch_async]

        # [START eventhub_producer_client_send_async]
        async with producer:
            event_data_batch = await producer.create_batch()
            while True:
                try:
                    event_data_batch.add(EventData("Message inside EventBatchData"))
                except ValueError:
                    # The EventDataBatch object reaches its max_size.
                    # You can send the full EventDataBatch object and create a new one here.
                    break
            await producer.send_batch(event_data_batch)
        # [END eventhub_producer_client_send_async]
        await asyncio.sleep(1)

        # [START eventhub_consumer_client_receive_async]
        logger = logging.getLogger("azure.eventhub")

        async def on_event(partition_context, event):
            # Put your code here.
            # If the operation is i/o intensive, async will have better performance.
            logger.info("Received event from partition: {}".format(partition_context.partition_id))

        async with consumer:
            await consumer.receive(
                on_event=on_event,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
        # [END eventhub_consumer_client_receive_async]

        consumer = example_create_async_eventhub_consumer_client()
        # [START eventhub_consumer_client_receive_batch_async]
        logger = logging.getLogger("azure.eventhub")

        async def on_event_batch(partition_context, event_batch):
            # Put your code here.
            # If the operation is i/o intensive, async will have better performance.
            logger.info(
                "{} events received from partition: {}".format(len(event_batch), partition_context.partition_id)
            )

        async with consumer:
            await consumer.receive_batch(
                on_event_batch=on_event_batch,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
        # [END eventhub_consumer_client_receive_batch_async]

    finally:
        pass


async def example_eventhub_async_producer_send_and_close():
    # [START eventhub_producer_client_close_async]
    import os
    from azure.eventhub.aio import EventHubProducerClient
    from azure.eventhub import EventData
    from azure.identity.aio import DefaultAzureCredential

    fully_qualified_namespace = os.environ["EVENT_HUB_HOSTNAME"]
    eventhub_name = os.environ["EVENT_HUB_NAME"]

    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,  # EventHub name should be specified if it doesn't show up in connection string.
        credential=DefaultAzureCredential(),
    )
    try:
        event_data_batch = await producer.create_batch()
        while True:
            try:
                event_data_batch.add(EventData("Message inside EventBatchData"))
            except ValueError:
                # The EventDataBatch object reaches its max_size.
                # You can send the full EventDataBatch object and create a new one here.
                break
        await producer.send_batch(event_data_batch)
    finally:
        # Close down the producer handler.
        await producer.close()
    # [END eventhub_producer_client_close_async]


async def example_eventhub_async_consumer_receive_and_close():
    # [START eventhub_consumer_client_close_async]
    import os
    from azure.identity.aio import DefaultAzureCredential

    fully_qualified_namespace = os.environ["EVENT_HUB_HOSTNAME"]
    eventhub_name = os.environ["EVENT_HUB_NAME"]

    from azure.eventhub.aio import EventHubConsumerClient

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        consumer_group="$Default",
        eventhub_name=eventhub_name,  # EventHub name should be specified if it doesn't show up in connection string.
        credential=DefaultAzureCredential(),
    )

    logger = logging.getLogger("azure.eventhub")

    async def on_event(partition_context, event):
        # Put your code here.
        # If the operation is i/o intensive, async will have better performance.
        logger.info("Received event from partition: {}".format(partition_context.partition_id))

    # The receive method is a coroutine which will be blocking when awaited.
    # It can be executed in an async task for non-blocking behavior, and combined with the 'close' method.

    recv_task = asyncio.ensure_future(consumer.receive(on_event=on_event))
    await asyncio.sleep(3)  # keep receiving for 3 seconds
    recv_task.cancel()  # stop receiving

    # Close down the consumer handler explicitly.
    await consumer.close()
    # [END eventhub_consumer_client_close_async]


if __name__ == "__main__":
    asyncio.run(example_eventhub_async_consumer_receive_and_close())
    asyncio.run(example_eventhub_async_producer_send_and_close())
    asyncio.run(example_eventhub_async_send_and_receive())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from typing import Any, Optional, List, Tuple


class DictMixin:
    def __setitem__(self, key: str, item: Any) -> None:
        self.__dict__[key] = item

    def __getitem__(self, key: str) -> Any:
        return self.__dict__[key]

    def __repr__(self) -> str:
        return str(self)

    def __len__(self) -> int:
        return len(self.keys())

    def __delitem__(self, key: str) -> None:
        self.__dict__[key] = None

    def __eq__(self, other: Any) -> bool:
        """Compare objects by comparing all attributes.

        :param any other: Any object
        :return: True if objects are equal, False otherwise.
        :rtype: bool
        """
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False

    def __ne__(self, other: Any) -> bool:
        """Compare objects by comparing all attributes.

        :param any other: Any object
        :return: True if objects are not equal, False otherwise.
        :rtype: bool
        """
        return not self.__eq__(other)

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def __str__(self) -> str:
        return str({k: v for k, v in self.__dict__.items() if not k.startswith("_")})

    def has_key(self, k: str) -> bool:
        return k in self.__dict__

    def update(self, *args: Any, **kwargs: Any) -> None:
        return self.__dict__.update(*args, **kwargs)

    def keys(self) -> List[str]:
        return [k for k in self.__dict__ if not k.startswith("_")]

    def values(self) -> List[Any]:
        return [v for k, v in self.__dict__.items() if not k.startswith("_")]

    def items(self) -> List[Tuple[str, Any]]:
        return [(k, v) for k, v in self.__dict__.items() if not k.startswith("_")]

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        if key in self.__dict__:
            return self.__dict__[key]
        return default



# -- coding: utf-8 --
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import platform
import pytest
import datetime
from packaging import version

try:
    import uamqp
    from azure.eventhub._transport._uamqp_transport import UamqpTransport
except (ModuleNotFoundError, ImportError):
    uamqp = None
    UamqpTransport = None
from azure.eventhub._transport._pyamqp_transport import PyamqpTransport
from azure.eventhub._pyamqp.message import Message, Properties, Header
from azure.eventhub._utils import CE_ZERO_SECONDS
from azure.eventhub._constants import PROP_TIMESTAMP
from azure.eventhub.amqp import AmqpAnnotatedMessage, AmqpMessageHeader, AmqpMessageProperties

from azure.eventhub import _common

pytestmark = pytest.mark.skipif(platform.python_implementation() == "PyPy", reason="This is ignored for PyPy")


from azure.eventhub import EventData, EventDataBatch


@pytest.mark.parametrize(
    "test_input, expected_result",
    [("", ""), ("AAA", "AAA"), (None, ValueError), (["a", "b", "c"], "abc"), (b"abc", "abc")],
)
def test_constructor(test_input, expected_result):
    if isinstance(expected_result, type):
        with pytest.raises(expected_result):
            EventData(test_input)
    else:
        event_data = EventData(test_input)
        assert event_data.body_as_str() == expected_result
        assert event_data.partition_key is None
        assert len(event_data.properties) == 0
        assert event_data.enqueued_time is None
        assert event_data.offset is None
        assert event_data.sequence_number is None
        assert len(event_data.system_properties) == 0
        assert str(event_data) == "{{ body: '{}', properties: {{}} }}".format(expected_result)
        assert repr(
            event_data
        ) == "EventData(body='{}', properties={{}}, offset=None, sequence_number=None, partition_key=None, enqueued_time=None)".format(
            expected_result
        )
        with pytest.raises(TypeError):
            event_data.body_as_json()


def test_body_json():
    event_data = EventData('{"a":"b"}')
    assert str(event_data) == '{ body: \'{"a":"b"}\', properties: {} }'
    assert (
        repr(event_data)
        == 'EventData(body=\'{"a":"b"}\', properties={}, offset=None, sequence_number=None, partition_key=None, enqueued_time=None)'
    )
    jo = event_data.body_as_json()
    assert jo["a"] == "b"


def test_body_wrong_json():
    event_data = EventData("aaa")
    with pytest.raises(TypeError):
        event_data.body_as_json()


def test_app_properties():
    app_props = {"a": "b"}
    event_data = EventData("")
    event_data.properties = app_props
    assert str(event_data) == "{ body: '', properties: {'a': 'b'} }"
    assert (
        repr(event_data)
        == "EventData(body='', properties={'a': 'b'}, offset=None, sequence_number=None, partition_key=None, enqueued_time=None)"
    )
    assert event_data.properties["a"] == "b"


def test_sys_properties(uamqp_transport):
    if uamqp_transport:
        properties = uamqp.message.MessageProperties()
        properties.message_id = "message_id"
        properties.user_id = "user_id"
        properties.to = "to"
        properties.subject = "subject"
        properties.reply_to = "reply_to"
        properties.correlation_id = "correlation_id"
        properties.content_type = "content_type"
        properties.content_encoding = "content_encoding"
        properties.absolute_expiry_time = 1
        properties.creation_time = 1
        properties.group_id = "group_id"
        properties.group_sequence = 1
        properties.reply_to_group_id = "reply_to_group_id"
        message = uamqp.message.Message(properties=properties)
        message.annotations = {_common.PROP_OFFSET: "@latest", PROP_TIMESTAMP: CE_ZERO_SECONDS * 1000}
    else:
        properties = Properties(
            message_id="message_id",
            user_id="user_id",
            to="to",
            subject="subject",
            reply_to="reply_to",
            correlation_id="correlation_id",
            content_type="content_type",
            content_encoding="content_encoding",
            absolute_expiry_time=1,
            creation_time=1,
            group_id="group_id",
            group_sequence=1,
            reply_to_group_id="reply_to_group_id",
        )
        message_annotations = {_common.PROP_OFFSET: "@latest", PROP_TIMESTAMP: CE_ZERO_SECONDS * 1000}
        message = Message(properties=properties, message_annotations=message_annotations)
    ed = EventData._from_message(message)  # type: EventData

    assert ed.system_properties[_common.PROP_OFFSET] == "@latest"
    assert ed.system_properties[_common.PROP_CORRELATION_ID] == properties.correlation_id
    assert ed.system_properties[_common.PROP_MESSAGE_ID] == properties.message_id
    assert ed.system_properties[_common.PROP_CONTENT_ENCODING] == properties.content_encoding
    assert ed.system_properties[_common.PROP_CONTENT_TYPE] == properties.content_type
    assert ed.system_properties[_common.PROP_USER_ID] == properties.user_id
    assert ed.system_properties[_common.PROP_TO] == properties.to
    assert ed.system_properties[_common.PROP_SUBJECT] == properties.subject
    assert ed.system_properties[_common.PROP_REPLY_TO] == properties.reply_to
    assert ed.system_properties[_common.PROP_ABSOLUTE_EXPIRY_TIME] == properties.absolute_expiry_time
    assert ed.system_properties[_common.PROP_CREATION_TIME] == properties.creation_time
    assert ed.system_properties[_common.PROP_GROUP_ID] == properties.group_id
    assert ed.system_properties[_common.PROP_GROUP_SEQUENCE] == properties.group_sequence
    assert ed.system_properties[_common.PROP_REPLY_TO_GROUP_ID] == properties.reply_to_group_id
    assert ed.enqueued_time == datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)


def test_event_data_batch(uamqp_transport):
    if uamqp_transport:
        if version.parse(uamqp.__version__) >= version.parse("1.2.8"):
            expected_result = 97
        else:
            expected_result = 93
        amqp_transport = UamqpTransport
    else:
        expected_result = 99
        amqp_transport = PyamqpTransport

    batch = EventDataBatch(max_size_in_bytes=110, partition_key="par", amqp_transport=amqp_transport)
    batch.add(EventData("A"))
    assert str(batch) == "EventDataBatch(max_size_in_bytes=110, partition_id=None, partition_key='par', event_count=1)"
    assert repr(batch) == "EventDataBatch(max_size_in_bytes=110, partition_id=None, partition_key='par', event_count=1)"

    assert batch.size_in_bytes == expected_result and len(batch) == 1

    with pytest.raises(ValueError):
        batch.add(EventData("A"))


def test_event_data_from_message(uamqp_transport):
    if uamqp_transport:
        amqp_transport = UamqpTransport
    else:
        amqp_transport = PyamqpTransport
    annotated_message = AmqpAnnotatedMessage(data_body=b"A")
    message = amqp_transport.to_outgoing_amqp_message(annotated_message)
    event = EventData._from_message(message)
    assert event.content_type is None
    assert event.correlation_id is None
    assert event.message_id is None

    event.content_type = "content_type"
    event.correlation_id = "correlation_id"
    event.message_id = "message_id"
    assert event.content_type == "content_type"
    assert event.correlation_id == "correlation_id"
    assert event.message_id == "message_id"
    assert list(event.body) == [b"A"]


def test_amqp_message_str_repr():
    data_body = b"A"
    message = AmqpAnnotatedMessage(data_body=data_body)
    assert str(message) == "A"
    assert "AmqpAnnotatedMessage(body=A, body_type=data" in repr(message)


def test_outgoing_amqp_message_header_properties(uamqp_transport):
    if uamqp_transport:
        amqp_transport = UamqpTransport
    else:
        amqp_transport = PyamqpTransport
    ann_message = AmqpAnnotatedMessage(data_body=b"A")
    ann_message.header = AmqpMessageHeader()
    ann_message.properties = AmqpMessageProperties()
    amqp_message = amqp_transport.to_outgoing_amqp_message(ann_message)

    assert not amqp_message.header
    assert not amqp_message.properties

    ann_message = AmqpAnnotatedMessage(data_body=b"A")
    ann_message.header = AmqpMessageHeader()
    ann_message.properties = AmqpMessageProperties()
    ann_message.header.first_acquirer = False
    ann_message.properties.creation_time = 0
    amqp_message = amqp_transport.to_outgoing_amqp_message(ann_message)

    assert amqp_message.header
    assert amqp_message.properties

    ann_message = AmqpAnnotatedMessage(data_body=b"A")
    ann_message.header = AmqpMessageHeader()
    ann_message.properties = AmqpMessageProperties()
    ann_message.properties.message_id = ""
    ann_message.application_properties = {b"key": b"val"}
    amqp_message = amqp_transport.to_outgoing_amqp_message(ann_message)

    assert not amqp_message.header
    assert amqp_message.properties
    assert isinstance(amqp_message.application_properties, dict)


def test_amqp_message_from_message(uamqp_transport):
    if uamqp_transport:
        header = uamqp.message.MessageHeader()
        header.delivery_count = 1
        header.time_to_live = 10000
        header.first_acquirer = True
        header.durable = True
        header.priority = 1
        properties = uamqp.message.MessageProperties()
        properties.message_id = "message_id"
        properties.user_id = "user_id"
        properties.to = "to"
        properties.subject = "subject"
        properties.reply_to = "reply_to"
        properties.correlation_id = "correlation_id"
        properties.content_type = "content_type"
        properties.content_encoding = "content_encoding"
        properties.absolute_expiry_time = 1
        properties.creation_time = 1
        properties.group_id = "group_id"
        properties.group_sequence = 1
        properties.reply_to_group_id = "reply_to_group_id"
        message = uamqp.message.Message(header=header, properties=properties)
        message.annotations = {_common.PROP_OFFSET: "@latest"}
    else:
        header = Header(delivery_count=1, ttl=10000, first_acquirer=True, durable=True, priority=1)
        properties = Properties(
            message_id="message_id",
            user_id="user_id",
            to="to",
            subject="subject",
            reply_to="reply_to",
            correlation_id="correlation_id",
            content_type="content_type",
            content_encoding="content_encoding",
            absolute_expiry_time=1,
            creation_time=1,
            group_id="group_id",
            group_sequence=1,
            reply_to_group_id="reply_to_group_id",
        )
        message_annotations = {_common.PROP_OFFSET: "@latest"}
        message = Message(properties=properties, header=header, message_annotations=message_annotations)

    amqp_message = AmqpAnnotatedMessage(message=message)
    assert amqp_message.properties.message_id == message.properties.message_id
    assert amqp_message.properties.user_id == message.properties.user_id
    assert amqp_message.properties.to == message.properties.to
    assert amqp_message.properties.subject == message.properties.subject
    assert amqp_message.properties.reply_to == message.properties.reply_to
    assert amqp_message.properties.correlation_id == message.properties.correlation_id
    assert amqp_message.properties.content_type == message.properties.content_type
    assert amqp_message.properties.absolute_expiry_time == message.properties.absolute_expiry_time
    assert amqp_message.properties.creation_time == message.properties.creation_time
    assert amqp_message.properties.group_id == message.properties.group_id
    assert amqp_message.properties.group_sequence == message.properties.group_sequence
    assert amqp_message.properties.reply_to_group_id == message.properties.reply_to_group_id
    assert amqp_message.header.time_to_live == message.header.ttl
    assert amqp_message.header.delivery_count == message.header.delivery_count
    assert amqp_message.header.first_acquirer == message.header.first_acquirer
    assert amqp_message.header.durable == message.header.durable
    assert amqp_message.header.priority == message.header.priority
    assert amqp_message.annotations == message.message_annotations


def test_legacy_message(uamqp_transport):
    if uamqp_transport:
        header = uamqp.message.MessageHeader()
        header.delivery_count = 1
        header.time_to_live = 10000
        header.first_acquirer = True
        header.durable = True
        header.priority = 1
        properties = uamqp.message.MessageProperties()
        properties.message_id = "message_id"
        properties.user_id = "user_id"
        properties.to = "to"
        properties.subject = "subject"
        properties.reply_to = "reply_to"
        properties.correlation_id = "correlation_id"
        properties.content_type = "content_type"
        properties.content_encoding = "content_encoding"
        properties.absolute_expiry_time = 1
        properties.creation_time = 1
        properties.group_id = "group_id"
        properties.group_sequence = 1
        properties.reply_to_group_id = "reply_to_group_id"
        message = uamqp.message.Message(body=b"abc", header=header, properties=properties)
        message.annotations = {_common.PROP_OFFSET: "@latest"}
        amqp_transport = UamqpTransport
    else:
        header = Header(delivery_count=1, ttl=10000, first_acquirer=True, durable=True, priority=1)
        properties = Properties(
            message_id="message_id",
            user_id="user_id",
            to="to",
            subject="subject",
            reply_to="reply_to",
            correlation_id="correlation_id",
            content_type="content_type",
            content_encoding="content_encoding",
            absolute_expiry_time=1,
            creation_time=1,
            group_id="group_id",
            group_sequence=1,
            reply_to_group_id="reply_to_group_id",
        )
        message_annotations = {_common.PROP_OFFSET: "@latest"}
        message = Message(data=b"abc", properties=properties, header=header, message_annotations=message_annotations)
        amqp_transport = PyamqpTransport
    event_data = EventData._from_message(message=message)
    assert event_data.message.properties.user_id == b"user_id"
    assert event_data.message.properties.message_id == b"message_id"
    assert event_data.message.properties.to == b"to"
    assert event_data.message.properties.subject == b"subject"
    assert event_data.message.properties.reply_to == b"reply_to"
    assert event_data.message.properties.correlation_id == b"correlation_id"
    assert event_data.message.properties.content_type == b"content_type"
    assert event_data.message.properties.content_encoding == b"content_encoding"
    assert event_data.message.properties.absolute_expiry_time == 1
    assert event_data.message.properties.creation_time == 1
    assert event_data.message.properties.group_id == b"group_id"
    assert event_data.message.properties.group_sequence == 1
    assert event_data.message.properties.reply_to_group_id == b"reply_to_group_id"
    assert event_data.message.state.value == 2
    assert event_data.message.delivery_annotations == {}
    assert event_data.message.delivery_no is None
    assert event_data.message.delivery_tag is None
    assert event_data.message.on_send_complete is None
    assert event_data.message.footer == {}
    assert event_data.message.retries == 0
    assert event_data.message.idle_time == 0

    event_data_batch = EventDataBatch(partition_key=b"par", partition_id="1", amqp_transport=amqp_transport)
    event_data_batch.add(event_data)
    assert event_data_batch.message.max_message_length == 1024 * 1024
    assert event_data_batch.message.size_offset == 0
    assert event_data_batch.message.batch_format == 0x80013700
    assert len(event_data_batch.message.annotations) == 1
    assert event_data_batch.message.application_properties is None
    assert event_data_batch.message.header.delivery_count == 0
    assert event_data_batch.message.header.time_to_live is None
    assert event_data_batch.message.header.first_acquirer is None
    assert event_data_batch.message.header.durable is True
    assert event_data_batch.message.header.priority is None
    assert event_data_batch.message.on_send_complete is None
    assert event_data_batch.message.properties is None

def test_from_bytes(uamqp_transport):
    if uamqp_transport:
        pytest.skip("This test is only for pyamqp transport")
    data = b'\x00Sr\xc1\x87\x08\xa3\x1bx-opt-sequence-number-epochT\xff\xa3\x15x-opt-sequence-numberU\x00\xa3\x0cx-opt-offsetU\x00\xa3\x13x-opt-enqueued-time\x00\xa3\x1dcom.microsoft:datetime-offset\x83\x00\x00\x01\x95\xf3XB\x86\x00St\xc1I\x02\xa1\rDiagnostic-Id\xa1700-1aa201483d464ac3c3d2ab796fbccb36-72e947bb22f404fc-00\x00Su\xa0\x08message1'

    event_data = EventData.from_bytes(data)
    assert event_data.body_as_str() == 'message1'
    assert event_data.sequence_number == 0
    assert event_data.offset is None
    assert event_data.properties == {b'Diagnostic-Id': b'00-1aa201483d464ac3c3d2ab796fbccb36-72e947bb22f404fc-00'}
    assert event_data.partition_key is None
    assert isinstance(event_data.enqueued_time, datetime.datetime)


# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import pytest
import time
from azure.identity import EnvironmentCredential, DefaultAzureCredential
from azure.eventhub import EventHubProducerClient, EventHubSharedKeyCredential
from azure.eventhub._client_base import EventHubSASTokenCredential
from azure.core.credentials import AzureSasCredential, AzureNamedKeyCredential


@pytest.mark.livetest
def test_mgmt_call_conn_str(connstr_receivers):
    connection_str, receivers = connstr_receivers
    client = EventHubProducerClient.from_connection_string(connection_str)
    client._start_producer("0", 60)


@pytest.mark.livetest
def test_mgmt_call_token_credential(live_eventhub, get_credential):
    credential = get_credential()
    client = EventHubProducerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=credential,
        user_agent="customized information",
    )
    client._start_producer("0", 60)


@pytest.mark.livetest
def test_mgmt_call_sas(live_eventhub):
    hostname = live_eventhub["hostname"]
    credential = EventHubSharedKeyCredential(live_eventhub["key_name"], live_eventhub["access_key"])
    auth_uri = "sb://{}/{}".format(hostname, live_eventhub["event_hub"])
    token = credential.get_token(auth_uri).token
    client = EventHubProducerClient(
        fully_qualified_namespace=hostname,
        eventhub_name=live_eventhub["event_hub"],
        credential=EventHubSASTokenCredential(token, time.time() + 3000),
    )
    client._start_producer("0", 60)
    assert True


@pytest.mark.livetest
def test_mgmt_call_sas_credential(live_eventhub):
    hostname = live_eventhub["hostname"]
    credential = EventHubSharedKeyCredential(live_eventhub["key_name"], live_eventhub["access_key"])
    auth_uri = "sb://{}/{}".format(hostname, live_eventhub["event_hub"])
    token = credential.get_token(auth_uri).token
    client = EventHubProducerClient(
        fully_qualified_namespace=hostname,
        eventhub_name=live_eventhub["event_hub"],
        credential=AzureSasCredential(token),
    )
    client._start_producer("0", 60)
    assert True


@pytest.mark.livetest
def test_mgmt_call_azure_named_key_credential(live_eventhub):
    credential = AzureNamedKeyCredential(live_eventhub["key_name"], live_eventhub["access_key"])
    client = EventHubProducerClient(
        fully_qualified_namespace=live_eventhub["hostname"],
        eventhub_name=live_eventhub["event_hub"],
        credential=credential,
    )

    client._start_producer("0", 60)
    assert True



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# -----------------------------------------------------------------------------------
from typing import Dict, Any, Iterable, Optional, Union
from azure.eventhub._eventprocessor.in_memory_checkpoint_store import (
    InMemoryCheckpointStore as CheckPointStoreImpl,
)
from .checkpoint_store import CheckpointStore


class InMemoryCheckpointStore(CheckpointStore):
    def __init__(self):
        self._checkpoint_store_impl = CheckPointStoreImpl()

    async def list_ownership(
        self, fully_qualified_namespace: str, eventhub_name: str, consumer_group: str, **kwargs: Any
    ) -> Iterable[Dict[str, Any]]:
        return self._checkpoint_store_impl.list_ownership(fully_qualified_namespace, eventhub_name, consumer_group)

    async def claim_ownership(
        self, ownership_list: Iterable[Dict[str, Any]], **kwargs: Any
    ) -> Iterable[Dict[str, Any]]:
        return self._checkpoint_store_impl.claim_ownership(ownership_list)

    async def update_checkpoint(self, checkpoint: Dict[str, Optional[Union[str, int]]], **kwargs: Any) -> None:
        self._checkpoint_store_impl.update_checkpoint(checkpoint)

    async def list_checkpoints(
        self, fully_qualified_namespace: str, eventhub_name: str, consumer_group: str, **kwargs: Any
    ) -> Iterable[Dict[str, Any]]:
        return self._checkpoint_store_impl.list_checkpoints(fully_qualified_namespace, eventhub_name, consumer_group)



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from an Event Hub partition with EventHubConsumerClient tracking
the last enqueued event properties of specific partition.
"""
import os
from azure.eventhub import EventHubConsumerClient
from azure.identity import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


def on_event(partition_context, event):
    print("Received event from partition {}.".format(partition_context.partition_id))

    # Put your code here. to do some operations on the event.
    # Avoid time-consuming operations.
    print(event)

    print(
        "Last enqueued event properties from partition: {} is: {}.".format(
            partition_context.partition_id, partition_context.last_enqueued_event_properties
        )
    )


if __name__ == "__main__":
    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        consumer_group="$Default",
        eventhub_name=EVENTHUB_NAME,
        credential=DefaultAzureCredential(),
    )

    try:
        with consumer_client:
            consumer_client.receive(
                on_event=on_event,
                partition_id="0",
                track_last_enqueued_event_properties=True,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
    except KeyboardInterrupt:
        print("Stopped receiving.")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import pytest
from azure.eventhub import (
    EventHubConnectionStringProperties,
    parse_connection_string,
)
from azure.eventhub._client_base import _parse_conn_str


class TestEventHubConnectionStringParser:

    def test_eh_conn_str_parse_cs(self, **kwargs):
        conn_str = "Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == "sb://eh-namespace.servicebus.windows.net/"
        assert parse_result.fully_qualified_namespace == "eh-namespace.servicebus.windows.net"
        assert parse_result.shared_access_key_name == "test-policy"
        assert parse_result.shared_access_key == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="

    def test_eh_conn_str_parse_with_entity_path(self, **kwargs):
        conn_str = "Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;EntityPath=eventhub-name"
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == "sb://eh-namespace.servicebus.windows.net/"
        assert parse_result.fully_qualified_namespace == "eh-namespace.servicebus.windows.net"
        assert parse_result.shared_access_key_name == "test-policy"
        assert parse_result.shared_access_key == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        assert parse_result.eventhub_name == "eventhub-name"

    def test_eh_conn_str_parse_sas_and_shared_key(self, **kwargs):
        conn_str = "Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;SharedAccessSignature=THISISASASXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == "Only one of the SharedAccessKey or SharedAccessSignature must be present."

    def test_eh_parse_malformed_conn_str_no_endpoint(self, **kwargs):
        conn_str = "SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == "Connection string is either blank or malformed."

    def test_eh_parse_malformed_conn_str_no_endpoint_value(self, **kwargs):
        conn_str = "Endpoint=;SharedAccessKeyName=test;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == "Connection string is either blank or malformed."

    def test_eh_parse_malformed_conn_str_no_netloc(self, **kwargs):
        conn_str = "Endpoint=MALFORMED;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == "Invalid Endpoint on the Connection String."

    def test_eh_parse_conn_str_sas(self, **kwargs):
        conn_str = "Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessSignature=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == "sb://eh-namespace.servicebus.windows.net/"
        assert parse_result.fully_qualified_namespace == "eh-namespace.servicebus.windows.net"
        assert parse_result.shared_access_signature == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        assert parse_result.shared_access_key_name == None

    def test_eh_parse_conn_str_whitespace_trailing_semicolon(self, **kwargs):
        conn_str = "    Endpoint=sb://resourcename.servicebus.windows.net/;SharedAccessSignature=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;    "
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == "sb://resourcename.servicebus.windows.net/"
        assert parse_result.fully_qualified_namespace == "resourcename.servicebus.windows.net"
        assert parse_result.shared_access_signature == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        assert parse_result.shared_access_key_name == None

    def test_eh_parse_conn_str_sas_trailing_semicolon(self, **kwargs):
        conn_str = "Endpoint=sb://resourcename.servicebus.windows.net/;SharedAccessSignature=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;"
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == "sb://resourcename.servicebus.windows.net/"
        assert parse_result.fully_qualified_namespace == "resourcename.servicebus.windows.net"
        assert parse_result.shared_access_signature == "THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        assert parse_result.shared_access_key_name == None

    def test_eh_parse_conn_str_no_keyname(self, **kwargs):
        conn_str = "Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert "Invalid connection string" in str(e.value)

    def test_eh_parse_conn_str_no_key(self, **kwargs):
        conn_str = "Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKeyName=test-policy"
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert "Invalid connection string" in str(e.value)

    def test_eh_parse_conn_str_no_key_or_sas(self, **kwargs):
        conn_str = "Endpoint=sb://resourcename.servicebus.windows.net/"
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == "At least one of the SharedAccessKey or SharedAccessSignature must be present."

    def test_eh_parse_malformed_conn_str_lowercase_endpoint(self, **kwargs):
        conn_str = "endpoint=sb://resourcename.servicebus.windows.net/;SharedAccessKeyName=test;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == "Connection string is either blank or malformed."

    def test_eh_parse_malformed_conn_str_lowercase_sa_key_name(self, **kwargs):
        conn_str = "Endpoint=sb://resourcename.servicebus.windows.net/;sharedaccesskeyname=test;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert "Invalid connection string" in str(e.value)

    def test_eh_parse_malformed_conn_str_lowercase_sa_key_name(self, **kwargs):
        conn_str = "Endpoint=sb://resourcename.servicebus.windows.net/;SharedAccessKeyName=test;sharedaccesskey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert "Invalid connection string" in str(e.value)

    def test_eh_emulator_slug_parse_localhost(self, **kwargs):
        conn_str = "Endpoint=sb://localhost:6065;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;UseDevelopmentEmulator=true"
        fully_qualified_namespace, policy, key, entity, signature, expiry, emulator = _parse_conn_str(conn_str)
        assert fully_qualified_namespace == "localhost:6065"
        assert emulator == True

    def test_eh_emulator_slug_parse_ipv6(self, **kwargs):
        conn_str = "Endpoint=sb://::1;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;UseDevelopmentEmulator=true"
        fully_qualified_namespace, policy, key, entity, signature, expiry, emulator = _parse_conn_str(conn_str)
        assert fully_qualified_namespace == "::1"
        assert emulator == True



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import List, TypedDict, Union, cast
from typing import Dict, Any
from tracing_common import FakeSpan
from devtools_testutils import (
    get_region_override,
    get_credential as get_devtools_credential,
)
from azure.eventhub.extensions.checkpointstoreblobaio import (
    BlobCheckpointStore as BlobCheckpointStoreAsync,
)
from azure.eventhub.extensions.checkpointstoreblob import BlobCheckpointStore
from azure.eventhub._pyamqp.authentication import SASTokenAuth
from azure.eventhub._pyamqp import ReceiveClient
from azure.eventhub import EventHubProducerClient, TransportType
from azure.mgmt.eventhub import EventHubManagementClient
from azure.core.settings import settings
from logging.handlers import RotatingFileHandler
from functools import partial
from typing import Callable, Literal
import sys
import os
import tempfile
import pytest
import logging
import uuid
import warnings
import subprocess
import time
import signal
import ssl
import functools

# Load environment variables from .env file automatically for all tests
from dotenv import load_dotenv

load_dotenv()


try:
    import uamqp

    uamqp_transport_params = [True, False]
    uamqp_transport_ids = ["uamqp", "pyamqp"]
except (ModuleNotFoundError, ImportError):
    uamqp_transport_params = [False]
    uamqp_transport_ids = ["pyamqp"]

socket_transports = [TransportType.Amqp, TransportType.AmqpOverWebsocket]
socket_transport_ids = ["amqp", "ws"]


collect_ignore = []
PARTITION_COUNT = 2
CONN_STR = "Endpoint=sb://{}/;SharedAccessKeyName={};SharedAccessKey={};EntityPath={}"
RES_GROUP_PREFIX = "eh-res-group"
NAMESPACE_PREFIX = "eh-ns"
EVENTHUB_PREFIX = "eh"
EVENTHUB_DEFAULT_AUTH_RULE_NAME = "RootManageSharedAccessKey"
LOCATION = get_region_override("westus")


@pytest.fixture(scope="function")
def faultinjector(live_eventhub, request: pytest.FixtureRequest):
    test_name = request.node.name
    faultinjector_params = cast(List[str], request.param["faultinjector_args"])

    if os.environ.get("USE_FAULTINJECTOR") != "true":
        pytest.skip(
            "Fault injector not enabled. See conftest.py::faultinjector for requirements."
        )
        yield
        return

    env_path = os.environ.get("FAULTINJECTOR_PATH")
    faultinjector_path = os.path.abspath(env_path) if env_path else None

    logs_dir = os.path.join(os.path.dirname(__file__), "faultinjector_logs", test_name)
    os.makedirs(logs_dir, exist_ok=True)

    print("Setting up fault injector")

    # Start the faultinjector process
    faultinjector_params.extend(
        ["--host", live_eventhub["hostname"], "--logs", logs_dir, "--cert", logs_dir]
    )

    print("fault injector arguments", *faultinjector_params)

    process = subprocess.Popen(
        [
            faultinjector_path,
            *faultinjector_params,
        ],
        stdout=sys.stdout,
        stderr=sys.stderr,
        preexec_fn=os.setsid,
    )

    if not process:
        raise RuntimeError(f"Failed to start faultinjector. See output for details.")

    try:
        time.sleep(1)

        yield AMQPPROXY_CLIENT_ARGS

        print(f"===> Fault injector logs are in '{logs_dir}'")

    finally:
        os.killpg(os.getpgid(process.pid), signal.SIGTERM)
        process.wait()


# Set up the amqpproxy environment variables
amqpproxy_path = os.environ.get("AMQPPROXY_PATH")
AMQPPROXY_PATH = os.path.abspath(amqpproxy_path) if amqpproxy_path else None
RECORD_AMQP_PROXY = os.environ.get("RECORD_AMQP_PROXY") == "true"
AMQPPROXY_RECORDINGS_DIR = os.path.join(
    os.path.dirname(__file__), "amqpproxy_recordings"
)
if RECORD_AMQP_PROXY:
    if not os.path.exists(AMQPPROXY_RECORDINGS_DIR):
        os.makedirs(AMQPPROXY_RECORDINGS_DIR)

    # Create/overwrite the amqpproxy startup log file
    AMQPPROXY_STARTUP_LOG = os.path.join(
        AMQPPROXY_RECORDINGS_DIR, "amqpproxy_startup.log"
    )
    if os.path.exists(AMQPPROXY_STARTUP_LOG):
        with open(AMQPPROXY_STARTUP_LOG, "w") as log_file:
            log_file.write("")  # Overwrite the file with an empty string
    else:
        # Create the file if it doesn't exist
        open(AMQPPROXY_STARTUP_LOG, "w").close()

AMQPPROXY_CUSTOM_ENDPOINT_ADDRESS = "sb://localhost:5671"
AMQPPROXY_TRANSPORT_TYPE = TransportType.Amqp

context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
context.check_hostname = False
context.verify_mode = ssl.CERT_NONE
AMQPPROXY_SSL_CONTEXT = context
AMQPPROXY_CLIENT_ARGS = {
    "custom_endpoint_address": AMQPPROXY_CUSTOM_ENDPOINT_ADDRESS,
    "ssl_context": AMQPPROXY_SSL_CONTEXT,
    "transport_type": AMQPPROXY_TRANSPORT_TYPE,
}


def pytest_addoption(parser):
    parser.addoption(
        "--sleep",
        action="store",
        default="True",
        help="sleep on reconnect test: True or False",
    )


@pytest.fixture
def sleep(request):
    sleep = request.config.getoption("--sleep")
    return sleep.lower() in ("true", "yes", "1", "y")


@pytest.fixture(scope="session", params=uamqp_transport_params, ids=uamqp_transport_ids)
def uamqp_transport(request):
    return request.param


@pytest.fixture(scope="session", params=socket_transports, ids=socket_transport_ids)
def socket_transport(request):
    return request.param


@pytest.fixture(scope="session")
def storage_url():
    try:
        account_name = os.environ["AZURE_STORAGE_ACCOUNT"]
        return f"https://{account_name}.blob.core.windows.net"
    except KeyError:
        warnings.warn(UserWarning("AZURE_STORAGE_ACCOUNT undefined"))
        pytest.skip("AZURE_STORAGE_ACCOUNT undefined")
        return


@pytest.fixture()
def checkpoint_store(storage_url):
    checkpoint_store = BlobCheckpointStore(
        storage_url,
        "blobcontainer" + str(uuid.uuid4()),
        credential=get_devtools_credential(),
    )
    return checkpoint_store


@pytest.fixture()
def checkpoint_store_aio(storage_url):
    checkpoint_store = BlobCheckpointStoreAsync(
        storage_url,
        "blobcontainer" + str(uuid.uuid4()),
        credential=get_devtools_credential(is_async=True),
    )
    return checkpoint_store


def get_logger(filename, level=logging.INFO):
    azure_logger = logging.getLogger("azure.eventhub")
    azure_logger.setLevel(level)
    uamqp_logger = logging.getLogger("uamqp")
    uamqp_logger.setLevel(logging.INFO)

    formatter = logging.Formatter("%(asctime)s %(name)-12s %(levelname)-8s %(message)s")
    console_handler = logging.StreamHandler(stream=sys.stdout)
    console_handler.setFormatter(formatter)
    if not azure_logger.handlers:
        azure_logger.addHandler(console_handler)
    if not uamqp_logger.handlers:
        uamqp_logger.addHandler(console_handler)

    if filename:
        file_handler = RotatingFileHandler(
            filename, maxBytes=5 * 1024 * 1024, backupCount=2
        )
        file_handler.setFormatter(formatter)
        azure_logger.addHandler(file_handler)

    return azure_logger


log = get_logger(None, logging.DEBUG)


@pytest.fixture(scope="session")
def timeout_factor(uamqp_transport):
    if uamqp_transport:
        return 1000
    else:
        return 1


@pytest.fixture(scope="session")
def fake_span():
    return FakeSpan


@pytest.fixture()
def enable_tracing(fake_span):
    settings.tracing_implementation.set_value(fake_span)
    yield fake_span
    # MUST RESET TRACING IMPLEMENTATION TO NONE, ELSE ALL OTHER TESTS ADD TRACING
    settings.tracing_implementation.set_value(None)


@pytest.fixture(scope="session")
def get_credential():
    return get_devtools_credential


@pytest.fixture(scope="session")
def get_credential_async():
    return partial(get_devtools_credential, is_async=True)


@pytest.fixture(scope="session")
def resource_group():
    try:
        resource_group_name = os.environ["EVENTHUB_RESOURCE_GROUP"]
    except KeyError:
        warnings.warn(UserWarning("EVENTHUB_RESOURCE_GROUP undefined - skipping test"))
        pytest.skip("EVENTHUB_RESOURCE_GROUP defined")
    return resource_group_name


@pytest.fixture(scope="session")
def eventhub_namespace(resource_group):
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        pytest.skip("AZURE_SUBSCRIPTION_ID defined")
        return
    base_url = os.environ.get(
        "EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/"
    )
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = EventHubManagementClient(
        get_devtools_credential(),
        SUBSCRIPTION_ID,
        base_url=base_url,
        credential_scopes=credential_scopes,
    )
    try:
        namespace_name = os.environ["EVENT_HUB_NAMESPACE"]
    except KeyError:
        warnings.warn(UserWarning("EVENT_HUB_NAMESPACE undefined - skipping test"))
        pytest.skip("EVENT_HUB_NAMESPACE defined")

    key = resource_client.namespaces.list_keys(
        resource_group, namespace_name, EVENTHUB_DEFAULT_AUTH_RULE_NAME
    )
    connection_string = key.primary_connection_string
    key_name = key.key_name
    primary_key = key.primary_key
    return namespace_name, connection_string, key_name, primary_key


@pytest.fixture()
def live_eventhub(
    resource_group, eventhub_namespace
):  # pylint: disable=redefined-outer-name
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        warnings.warn(UserWarning("AZURE_SUBSCRIPTION_ID undefined - skipping test"))
        pytest.skip("AZURE_SUBSCRIPTION_ID defined")

    base_url = os.environ.get(
        "EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/"
    )
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = EventHubManagementClient(
        get_devtools_credential(),
        SUBSCRIPTION_ID,
        base_url=base_url,
        credential_scopes=credential_scopes,
    )
    eventhub_name = EVENTHUB_PREFIX + str(uuid.uuid4())
    eventhub_ns_name, connection_string, key_name, primary_key = eventhub_namespace
    eventhub_endpoint_suffix = os.environ.get(
        "EVENT_HUB_ENDPOINT_SUFFIX", ".servicebus.windows.net"
    )
    try:
        eventhub = resource_client.event_hubs.create_or_update(
            resource_group,
            eventhub_ns_name,
            eventhub_name,
            {"partition_count": PARTITION_COUNT},
        )
        live_eventhub_config = {
            "resource_group": resource_group,
            "hostname": "{}{}".format(eventhub_ns_name, eventhub_endpoint_suffix),
            "key_name": key_name,
            "access_key": primary_key,
            "namespace": eventhub_ns_name,
            "event_hub": eventhub.name,
            "consumer_group": "$Default",
            "partition": "0",
            "connection_str": connection_string + ";EntityPath=" + eventhub.name,
        }
        yield live_eventhub_config
    finally:
        try:
            resource_client.event_hubs.delete(
                resource_group, eventhub_ns_name, eventhub_name
            )
        except:
            warnings.warn(UserWarning("eventhub teardown failed"))


@pytest.fixture()
def resource_mgmt_client():
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        warnings.warn(UserWarning("AZURE_SUBSCRIPTION_ID undefined - skipping test"))
        pytest.skip("AZURE_SUBSCRIPTION_ID defined")

    base_url = os.environ.get(
        "EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/"
    )
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = EventHubManagementClient(
        get_devtools_credential(),
        SUBSCRIPTION_ID,
        base_url=base_url,
        credential_scopes=credential_scopes,
    )
    yield resource_client


@pytest.fixture()
def connection_str(live_eventhub):
    return CONN_STR.format(
        live_eventhub["hostname"],
        live_eventhub["key_name"],
        live_eventhub["access_key"],
        live_eventhub["event_hub"],
    )


@pytest.fixture()
def skip_amqp_proxy(request):
    """Helper method to determine if the AMQP proxy should be run for a test."""
    if not RECORD_AMQP_PROXY or not AMQPPROXY_PATH:
        return True
    if any(marker.name == "no_amqpproxy" for marker in request.node.own_markers):
        return True
    return False


@pytest.fixture()
def client_args(skip_amqp_proxy):
    """Fixture that adds the amqpproxy client args to the test context."""
    if skip_amqp_proxy:
        return {}
    # Add proxy args to test context
    return AMQPPROXY_CLIENT_ARGS


def remove_existing_recordings(path, file_name):
    """Remove existing recordings for the test."""
    # Remove any existing recordings for the test
    for file in os.listdir(path):
        if file.startswith(file_name) and file.endswith(".json"):
            os.remove(os.path.join(path, file))
            print(f"Removed existing recording: {file}")


def stop_existing_amqpproxy(log_file):
    # Kill any existing process using the AMQP proxy port
    try:
        subprocess.run(
            ["fuser", "-k", "5671/tcp"], check=True, stdout=log_file, stderr=log_file
        )
        log_file.write("Kill existing process on port 5671.\n")
    except subprocess.CalledProcessError:
        log_file.write("No existing process found on port 5671.\n")


@pytest.fixture(autouse=True)
def amqpproxy(live_eventhub, skip_amqp_proxy, request):
    """Fixture that redirects network requests to target the amqp proxy.
    Tests can opt out using @pytest.mark.no_amqpproxy or set the environment variable
    RECORD_AMQP_PROXY=False.
    """
    # Skip if not recording or test opted out
    if skip_amqp_proxy:
        yield
        return

    # Use test name as logfile
    test_name = request.node.name
    # Mirror relative path in AMQPPROXY_RECORDINGS_PATH for recording files
    relative_path = os.path.relpath(
        request.node.fspath, start=os.path.dirname(__file__)
    )
    recording_dir_path = os.path.join(
        AMQPPROXY_RECORDINGS_DIR, os.path.dirname(relative_path)
    )
    file_name = os.path.splitext(os.path.basename(request.node.fspath))[0]
    recording_file = f"{file_name}.{test_name}"
    if not os.path.exists(recording_dir_path):
        os.makedirs(recording_dir_path)
    else:
        # Remove any existing recordings with the same test name, so that we overwrite instead of add
        remove_existing_recordings(recording_dir_path, recording_file)

    # Start amqpproxy process
    log_file = open(AMQPPROXY_STARTUP_LOG, "a")
    # Flush log after writing to ensure log line ordering is preserved
    log_file.write(f"####### Starting amqpproxy for test: {test_name}\n")
    log_file.flush()
    try:
        # Navigate to the amqpproxy directory and run the proxy
        os.chdir(AMQPPROXY_PATH)
        stop_existing_amqpproxy(log_file)

        # Start the AMQP proxy process
        log_file.write("Starting amqpproxy process...\n")
        log_file.flush()
        proxy_process = subprocess.Popen(
            [
                "go",
                "run",
                ".",
                "--host",
                live_eventhub["hostname"],
                "--logs",
                recording_dir_path,
                "--logfile",
                recording_file,
            ],
            stdout=log_file,
            stderr=log_file,
            preexec_fn=os.setsid,
        )

        if not proxy_process:
            log_file.write("Failed to start amqpproxy.\n")
            raise RuntimeError(
                f"Failed to start amqpproxy. Check for errors in {AMQPPROXY_STARTUP_LOG}"
            )

        try:
            time.sleep(1)
            # Add proxy args to test context
            request.node.user_properties.append(("client_args", AMQPPROXY_CLIENT_ARGS))
            yield
        finally:
            os.killpg(os.getpgid(proxy_process.pid), signal.SIGTERM)
            proxy_process.wait()
    finally:
        log_file.write(f"####### Stopping amqpproxy for test: {test_name}\n")
        log_file.flush()
        log_file.close()


@pytest.fixture()
def invalid_hostname(live_eventhub):
    return CONN_STR.format(
        "invalid123.servicebus.windows.net",
        live_eventhub["key_name"],
        live_eventhub["access_key"],
        live_eventhub["event_hub"],
    )


@pytest.fixture()
def invalid_key(live_eventhub):
    return CONN_STR.format(
        live_eventhub["hostname"],
        live_eventhub["key_name"],
        "invalid",
        live_eventhub["event_hub"],
    )


@pytest.fixture()
def invalid_policy(live_eventhub):
    return CONN_STR.format(
        live_eventhub["hostname"],
        "invalid",
        live_eventhub["access_key"],
        live_eventhub["event_hub"],
    )


@pytest.fixture()
def connstr_receivers(live_eventhub, uamqp_transport, client_args):
    connection_str = live_eventhub["connection_str"]
    partitions = [str(i) for i in range(PARTITION_COUNT)]
    receivers = []
    for p in partitions:
        uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
        source = "amqps://{}/{}/ConsumerGroups/{}/Partitions/{}".format(
            live_eventhub["hostname"],
            live_eventhub["event_hub"],
            live_eventhub["consumer_group"],
            p,
        )
        if uamqp_transport:
            sas_auth = uamqp.authentication.SASTokenAuth.from_shared_access_key(
                uri, live_eventhub["key_name"], live_eventhub["access_key"]
            )
            receiver = uamqp.ReceiveClient(
                source, auth=sas_auth, debug=False, timeout=0, prefetch=500
            )
        else:
            sas_auth = SASTokenAuth(
                uri, uri, live_eventhub["key_name"], live_eventhub["access_key"]
            )
            receiver = ReceiveClient(
                live_eventhub["hostname"],
                source,
                auth=sas_auth,
                network_trace=False,
                timeout=0,
                link_credit=500,
                **client_args,
            )
        receiver.open()
        receivers.append(receiver)
    yield connection_str, receivers
    for r in receivers:
        r.close()


@pytest.fixture()
def auth_credentials(live_eventhub):
    return (
        live_eventhub["hostname"],
        live_eventhub["event_hub"],
        get_devtools_credential,
    )


@pytest.fixture()
def auth_credentials_async(live_eventhub):
    return (
        live_eventhub["hostname"],
        live_eventhub["event_hub"],
        partial(get_devtools_credential, is_async=True),
    )


@pytest.fixture()
def auth_credential_receivers(live_eventhub, uamqp_transport, client_args):
    fully_qualified_namespace = live_eventhub["hostname"]
    eventhub_name = live_eventhub["event_hub"]
    partitions = [str(i) for i in range(PARTITION_COUNT)]
    receivers = []
    for p in partitions:
        uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
        source = "amqps://{}/{}/ConsumerGroups/{}/Partitions/{}".format(
            live_eventhub["hostname"],
            live_eventhub["event_hub"],
            live_eventhub["consumer_group"],
            p,
        )
        if uamqp_transport:
            sas_auth = uamqp.authentication.SASTokenAuth.from_shared_access_key(
                uri, live_eventhub["key_name"], live_eventhub["access_key"]
            )
            receiver = uamqp.ReceiveClient(
                source, auth=sas_auth, debug=False, timeout=0, prefetch=500
            )
        else:
            # TODO: TokenAuth should be fine?
            sas_auth = SASTokenAuth(
                uri, uri, live_eventhub["key_name"], live_eventhub["access_key"]
            )
            receiver = ReceiveClient(
                live_eventhub["hostname"],
                source,
                auth=sas_auth,
                network_trace=False,
                timeout=30,
                link_credit=500,
                **client_args,
            )
        receiver.open()
        receivers.append(receiver)
    yield fully_qualified_namespace, eventhub_name, get_devtools_credential, receivers
    for r in receivers:
        r.close()


@pytest.fixture()
def auth_credential_receivers_async(live_eventhub, uamqp_transport, client_args):
    fully_qualified_namespace = live_eventhub["hostname"]
    eventhub_name = live_eventhub["event_hub"]
    partitions = [str(i) for i in range(PARTITION_COUNT)]
    receivers = []
    for p in partitions:
        uri = "sb://{}/{}".format(live_eventhub["hostname"], live_eventhub["event_hub"])
        source = "amqps://{}/{}/ConsumerGroups/{}/Partitions/{}".format(
            live_eventhub["hostname"],
            live_eventhub["event_hub"],
            live_eventhub["consumer_group"],
            p,
        )
        if uamqp_transport:
            sas_auth = uamqp.authentication.SASTokenAuth.from_shared_access_key(
                uri, live_eventhub["key_name"], live_eventhub["access_key"]
            )
            receiver = uamqp.ReceiveClient(
                source, auth=sas_auth, debug=False, timeout=0, prefetch=500
            )
        else:
            # TODO: TokenAuth should be fine?
            sas_auth = SASTokenAuth(
                uri, uri, live_eventhub["key_name"], live_eventhub["access_key"]
            )
            receiver = ReceiveClient(
                live_eventhub["hostname"],
                source,
                auth=sas_auth,
                network_trace=False,
                timeout=30,
                link_credit=500,
                **client_args,
            )
        receiver.open()
        receivers.append(receiver)
    yield fully_qualified_namespace, eventhub_name, partial(
        get_devtools_credential, is_async=True
    ), receivers
    for r in receivers:
        r.close()


@pytest.fixture()
def auth_credential_senders(live_eventhub, uamqp_transport, client_args):
    fully_qualified_namespace = live_eventhub["hostname"]
    eventhub_name = live_eventhub["event_hub"]
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=get_devtools_credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    partitions = client.get_partition_ids()

    senders = []
    for p in partitions:
        sender = client._create_producer(partition_id=p)
        senders.append(sender)
    yield fully_qualified_namespace, eventhub_name, get_devtools_credential, senders
    for s in senders:
        s.close()
    client.close()


@pytest.fixture()
def auth_credential_senders_async(live_eventhub, uamqp_transport, client_args):
    fully_qualified_namespace = live_eventhub["hostname"]
    eventhub_name = live_eventhub["event_hub"]
    client = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=get_devtools_credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    partitions = client.get_partition_ids()

    senders = []
    for p in partitions:
        sender = client._create_producer(partition_id=p)
        senders.append(sender)
    yield fully_qualified_namespace, eventhub_name, partial(
        get_devtools_credential, is_async=True
    ), senders
    for s in senders:
        s.close()
    client.close()


# Note: This is duplicated between here and the basic conftest, so that it does not throw warnings if you're
# running locally to this SDK. (Everything works properly, pytest just makes a bit of noise.)
def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line("markers", "liveTest: mark test to be a live test only")
    config.addinivalue_line(
        "markers", "no_amqpproxy: mark test to opt out of amqp proxy recording"
    )



import asyncio
import sys

from azure.eventhub._pyamqp.error import AMQPLinkError

if sys.version_info >= (3, 8, 0):
    from unittest.mock import AsyncMock
from azure.eventhub._pyamqp.aio import Link
from azure.eventhub._pyamqp.constants import LinkState
import pytest


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "start_state,expected_state",
    [
        (LinkState.ATTACHED, LinkState.DETACH_SENT),
        (LinkState.ATTACH_SENT, LinkState.DETACHED),
        (LinkState.ATTACH_RCVD, LinkState.DETACHED),
    ],
    ids=["link attached", "link attach sent", "link attach rcvd"],
)
async def test_link_should_detach(start_state, expected_state):
    if sys.version_info < (3, 8, 0):
        pytest.skip("AsyncMock is not available in Python 3.7")
    session = AsyncMock()
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    assert link.state == LinkState.DETACHED

    await link._set_state(start_state)
    link._outgoing_detach = AsyncMock(return_value=None)
    await link.detach()

    assert link.state == expected_state


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "state",
    [LinkState.DETACHED, LinkState.DETACH_SENT, LinkState.ERROR],
    ids=["link detached", "link detach sent", "link error"],
)
async def test_link_should_not_detach(state):
    if sys.version_info < (3, 8, 0):
        pytest.skip("AsyncMock is not available in Python 3.7")
    session = None
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    assert link.state == LinkState.DETACHED

    await link._set_state(state)
    link._outgoing_detach = AsyncMock(return_value=None)
    await link.detach()
    link._outgoing_detach.assert_not_called()

@pytest.mark.asyncio
@pytest.mark.parametrize(
    "frame",
    [
        [2, True, [b'amqp:link:detach-forced', b"The link is force detached. Code: publisher(link3006875). Details: AmqpMessagePublisher.IdleTimerExpired: Idle timeout: 00:10:00.", None]],
        [2, True, [b'amqp:link:detach-forced', None, b'something random']],
        [2, True, [b'amqp:link:detach-forced', None, None]],
        [2, True, [b'amqp:link:detach-forced']],
        
    ],
    ids=["description and info", "info only", "description only", "no info or description"],
)
async def test_detach_with_error(frame):
    '''
      A detach can optionally include an description and info field.
      https://docs.oasis-open.org/amqp/core/v1.0/os/amqp-core-transport-v1.0-os.html#type-error
    '''
    session = None
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    await link._set_state(LinkState.DETACH_RCVD)
    await link._incoming_detach(frame)

    with pytest.raises(AMQPLinkError) as ae:
        await link.get_state()
        assert ae.description == frame[2][1]
        assert ae.info == frame[2][2]


# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import sys
import pytest

from azure.eventhub import EventHubSharedKeyCredential
from azure.eventhub import EventHubConsumerClient
from azure.eventhub.exceptions import (
    AuthenticationError,
    ConnectError,
)


@pytest.mark.liveTest
def test_get_properties(auth_credentials, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        consumer_group="$default",
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    with client:
        properties = client.get_eventhub_properties()
        assert properties["eventhub_name"] == eventhub_name and properties["partition_ids"] == ["0", "1"]


@pytest.mark.liveTest
def test_get_properties_with_auth_error_sync(auth_credentials, live_eventhub, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, _ = auth_credentials
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        consumer_group="$default",
        credential=EventHubSharedKeyCredential(live_eventhub["key_name"], "AaBbCcDdEeFf="),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    with client:
        with pytest.raises(AuthenticationError) as e:
            client.get_eventhub_properties()

    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        consumer_group="$default",
        credential=EventHubSharedKeyCredential("invalid", live_eventhub["access_key"]),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    with client:
        with pytest.raises(AuthenticationError) as e:
            client.get_eventhub_properties()


@pytest.mark.liveTest
def test_get_properties_with_connect_error(auth_credentials, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name="invalid",
        consumer_group="$default",
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    with client:
        with pytest.raises(ConnectError) as e:
            client.get_eventhub_properties()

    client = EventHubConsumerClient(
        fully_qualified_namespace="invalid.servicebus.windows.net",
        eventhub_name=eventhub_name,
        consumer_group="$default",
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    with client:
        with pytest.raises(ConnectError) as e:
            client.get_eventhub_properties()


@pytest.mark.liveTest
def test_get_partition_ids(auth_credentials, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        consumer_group="$default",
        credential=credential(),
        uamqp_transport=uamqp_transport,
        **client_args,
    )
    with client:
        partition_ids = client.get_partition_ids()
        assert partition_ids == ["0", "1"]


@pytest.mark.liveTest
def test_get_partition_properties(auth_credentials, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials
    client = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        consumer_group="$default",
        credential=credential(),
        uamqp_transport=uamqp_transport,
        client_args=client_args,
    )
    with client:
        properties = client.get_partition_properties("0")
        assert (
            properties["eventhub_name"] == eventhub_name
            and properties["id"] == "0"
            and "beginning_sequence_number" in properties
            and "last_enqueued_sequence_number" in properties
            and "last_enqueued_offset" in properties
            and "last_enqueued_time_utc" in properties
            and "is_empty" in properties
        )



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import logging
import uuid
import time
from functools import partial

from ._management_link_async import ManagementLink
from ..error import AMQPLinkError, ErrorCondition

from ..constants import ManagementOpenResult, ManagementExecuteOperationResult

_LOGGER = logging.getLogger(__name__)


class ManagementOperation(object):
    def __init__(self, session, endpoint="$management", **kwargs):
        self._mgmt_link_open_status = None

        self._session = session
        self._connection = self._session._connection
        self._network_trace_params = {
            "amqpConnection": self._session._connection._container_id,
            "amqpSession": self._session.name,
            "amqpLink": "",
        }
        self._mgmt_link: ManagementLink = self._session.create_request_response_link_pair(
            endpoint=endpoint,
            on_amqp_management_open_complete=self._on_amqp_management_open_complete,
            on_amqp_management_error=self._on_amqp_management_error,
            **kwargs
        )
        self._responses = {}
        self._mgmt_error = None

    async def _on_amqp_management_open_complete(self, result):
        """Callback run when the send/receive links are open and ready
        to process messages.

        :param result: Whether the link opening was successful.
        :type result: int
        """
        self._mgmt_link_open_status = result

    async def _on_amqp_management_error(self):
        """Callback run if an error occurs in the send/receive links."""
        # TODO: This probably shouldn't be ValueError
        self._mgmt_error = ValueError("Management Operation error occurred.")

    async def _on_execute_operation_complete(
        self, operation_id, operation_result, status_code, status_description, raw_message, error=None
    ):
        _LOGGER.debug(
            "Management operation completed, id: %r; result: %r; code: %r; description: %r, error: %r",
            operation_id,
            operation_result,
            status_code,
            status_description,
            error,
            extra=self._network_trace_params,
        )

        if operation_result in (ManagementExecuteOperationResult.ERROR, ManagementExecuteOperationResult.LINK_CLOSED):
            self._mgmt_error = error
            _LOGGER.error(
                "Failed to complete management operation due to error: %r.", error, extra=self._network_trace_params
            )
        else:
            self._responses[operation_id] = (status_code, status_description, raw_message)

    async def execute(self, message, operation=None, operation_type=None, timeout: float = 0):
        start_time = time.time()
        operation_id = str(uuid.uuid4())
        self._responses[operation_id] = None
        self._mgmt_error = None

        await self._mgmt_link.execute_operation(
            message,
            partial(self._on_execute_operation_complete, operation_id),
            timeout=timeout,
            operation=operation,
            type=operation_type,
        )

        while not self._responses[operation_id] and not self._mgmt_error:
            if timeout and timeout > 0:
                now = time.time()
                if (now - start_time) >= timeout:
                    raise TimeoutError("Failed to receive mgmt response in {}ms".format(timeout))
            await self._connection.listen()

        if self._mgmt_error:
            self._responses.pop(operation_id)
            raise self._mgmt_error

        response = self._responses.pop(operation_id)
        return response

    async def open(self):
        self._mgmt_link_open_status = ManagementOpenResult.OPENING
        await self._mgmt_link.open()

    async def ready(self):
        try:
            raise self._mgmt_error
        except TypeError:
            pass

        if self._mgmt_link_open_status == ManagementOpenResult.OPENING:
            return False
        if self._mgmt_link_open_status == ManagementOpenResult.OK:
            return True
        # ManagementOpenResult.ERROR or CANCELLED
        # TODO: update below with correct status code + info
        raise AMQPLinkError(
            condition=ErrorCondition.ClientError,
            description="Failed to open mgmt link, management link status: {}".format(self._mgmt_link_open_status),
            info=None,
        )

    async def close(self):
        await self._mgmt_link.close()



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
"""
An example to show authentication using a connection string obtained via the Azure Portal, or the Azure CLI toolkit.
"""

import asyncio
import os
from azure.eventhub.aio import EventHubConsumerClient

CONNECTION_STR = os.environ["EVENT_HUB_CONN_STR"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


async def main():
    consumer_client = EventHubConsumerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        consumer_group="$Default",
        eventhub_name=EVENTHUB_NAME,
    )
    async with consumer_client:
        pass  # consumer_client is now ready to be used.


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Examples to show sending events in buffered mode to an Event Hub.
"""

import time
import os
from azure.eventhub import EventHubProducerClient, EventData
from azure.identity import DefaultAzureCredential

FULLY_QUALIFIED_NAMESPACE = os.environ["EVENT_HUB_HOSTNAME"]
EVENTHUB_NAME = os.environ["EVENT_HUB_NAME"]


def on_success(events, pid):
    # sending succeeded
    print(events, pid)


def on_error(events, pid, error):
    # sending failed
    print(events, pid, error)


producer = EventHubProducerClient(
    fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
    eventhub_name=EVENTHUB_NAME,
    credential=DefaultAzureCredential(),
    buffered_mode=True,
    on_success=on_success,
    on_error=on_error,
)

start_time = time.time()

# exiting the context manager will automatically call flush
with producer:
    # single events will be batched automatically
    for i in range(10):
        # the method returning indicates the event has been enqueued to the buffer
        producer.send_event(EventData("Single data {}".format(i)))

    batch = producer.create_batch()
    for i in range(10):
        batch.add(EventData("Single data in batch {}".format(i)))
    # alternatively, you can enqueue an EventDataBatch object to the buffer
    producer.send_batch(batch)

    # calling flush sends out the events in the buffer immediately
    producer.flush()

print("Send messages in {} seconds.".format(time.time() - start_time))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from ._buffered_producer_dispatcher_async import BufferedProducerDispatcher
from ._partition_resolver_async import PartitionResolver
from ._buffered_producer_async import BufferedProducer

__all__ = ["BufferedProducerDispatcher", "PartitionResolver", "BufferedProducer"]



import pytest
from azure.eventhub._pyamqp._decode import _decode_decimal128
from decimal import Decimal



@pytest.mark.parametrize(
    "value,expected",
    [
        (b'\x0c>\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"\xa1', Decimal("3.4")),
        (b'\x0c6\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x86\xb9\xa1', Decimal(".34489")),
        (b'\x0c@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\n\xa1', Decimal("10")),
        (b'\x0c8\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x14%\xd8)\xa1', Decimal("33802.4489")),
    ]
)
def test_decimal_decode(value, expected):
    output = _decode_decimal128(memoryview(value))
    assert output[1] == expected





#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import asyncio
from collections import defaultdict
from uuid import uuid4

import pytest

from azure.eventhub import EventData
from azure.eventhub.aio import EventHubProducerClient, EventHubConsumerClient
from azure.eventhub.aio._buffered_producer import PartitionResolver
from azure.eventhub.amqp import (
    AmqpAnnotatedMessage,
)
from azure.eventhub.exceptions import (
    EventDataSendError,
    OperationTimeoutError,
    EventHubError,
)


async def random_pkey_generation(partitions):
    pr = PartitionResolver(partitions)
    total = len(partitions)
    dic = {}

    while total:
        key = str(uuid4())
        pid = await pr.get_partition_id_by_partition_key(key)
        if pid in dic:
            continue
        else:
            dic[pid] = key
            total -= 1

    return dic


@pytest.mark.liveTest()
@pytest.mark.asyncio
async def test_producer_client_constructor(auth_credentials_async, uamqp_transport, client_args):
    async def on_success(events, pid):
        pass

    async def on_error(events, error, pid):
        pass

    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    with pytest.raises(TypeError):
        EventHubProducerClient(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=credential(),
            buffered_mode=True,
            uamqp_transport=uamqp_transport,
            **client_args
        )
    with pytest.raises(TypeError):
        EventHubProducerClient(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=credential(),
            buffered_mode=True,
            on_success=on_success,
            uamqp_transport=uamqp_transport,
            **client_args
        )
    with pytest.raises(TypeError):
        EventHubProducerClient(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=credential(),
            buffered_mode=True,
            on_error=on_error,
            uamqp_transport=uamqp_transport,
            **client_args
        )
    with pytest.raises(ValueError):
        EventHubProducerClient(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=credential(),
            buffered_mode=True,
            on_success=on_success,
            on_error=on_error,
            max_wait_time=0,
            uamqp_transport=uamqp_transport,
            **client_args
        )
    with pytest.raises(ValueError):
        EventHubProducerClient(
            fully_qualified_namespace=fully_qualified_namespace,
            eventhub_name=eventhub_name,
            credential=credential(),
            buffered_mode=True,
            on_success=on_success,
            on_error=on_error,
            max_buffer_length=0,
            uamqp_transport=uamqp_transport,
            **client_args
        )

    def on_success_missing_params(events):
        on_success_missing_params.events = events

    def on_error_missing_params(events, pid):
        on_error_missing_params.events = events

    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        buffered_mode=True,
        buffer_concurrency=2,
        on_success=on_success_missing_params,
        on_error=on_error_missing_params,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    on_success_missing_params.events = None
    on_error_missing_params.events = None

    # successfully send, but don't enter invalid callback
    async with producer:
        await producer.send_event(EventData("Single data"))

    assert not on_success_missing_params.events
    assert not on_error_missing_params.events


@pytest.mark.liveTest
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "flush_after_sending, close_after_sending",
    [(False, False), (True, False), (False, True)],
)
async def test_basic_send_single_events_round_robin(
    auth_credentials_async, flush_after_sending, close_after_sending, uamqp_transport, client_args
):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    received_events = defaultdict(list)

    async def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )
    receive_thread = asyncio.ensure_future(consumer.receive(on_event=on_event))

    await asyncio.sleep(5)
    sent_events = defaultdict(list)

    async def on_success(events, pid):
        if len(events) > 1:
            on_success.batching = True
        sent_events[pid].extend(events)

    async def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None  # ensure no error
    on_success.batching = False  # ensure batching happened
    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer:
        partitions = await producer.get_partition_ids()
        partitions_cnt = len(partitions)
        # perform single sending round-robin
        total_single_event_cnt = 100
        eventdata_set, amqpannoated_set = set(), set()
        for i in range(total_single_event_cnt // 2):
            event = EventData("test:{}".format(i))
            event.properties = {"event_idx": i}
            await producer.send_event(event)
            eventdata_set.add(i)
        for i in range(total_single_event_cnt // 2, total_single_event_cnt):
            event = AmqpAnnotatedMessage(data_body="test:{}".format(i))
            event.application_properties = {"event_idx": i}
            amqpannoated_set.add(i)
            await producer.send_event(event)

        for pid in partitions:
            assert producer.get_buffered_event_count(pid) > 0
        assert producer.total_buffered_event_count > 0

        if not flush_after_sending and not close_after_sending:
            # ensure it's buffered sending
            for pid in partitions:
                assert len(sent_events[pid]) < total_single_event_cnt // partitions_cnt
            assert sum([len(sent_events[pid]) for pid in partitions]) < total_single_event_cnt
        else:
            if flush_after_sending:
                await producer.flush()
            if close_after_sending:
                await producer.close()
            # ensure all events are sent after calling flush
            assert sum([len(sent_events[pid]) for pid in partitions]) == total_single_event_cnt

        # give some time for producer to complete sending and consumer to complete receiving
        await asyncio.sleep(10)
        assert len(sent_events) == len(received_events) == partitions_cnt

        for pid in partitions:
            assert producer.get_buffered_event_count(pid) == 0
        assert producer.total_buffered_event_count == 0
        assert not on_error.err

        # ensure all events are received in the correct partition
        for pid in partitions:
            assert len(sent_events[pid]) >= total_single_event_cnt // partitions_cnt
            assert len(sent_events[pid]) == len(received_events[pid])
            for i in range(len(sent_events[pid])):
                event = sent_events[pid][i]
                try:  # amqp annotated message
                    event_idx = event.application_properties["event_idx"]
                    amqpannoated_set.remove(event_idx)
                except AttributeError:  # event data
                    event_idx = event.properties["event_idx"]
                    eventdata_set.remove(event_idx)
                assert received_events[pid][i].properties[b"event_idx"] == event_idx
                assert partitions[event_idx % partitions_cnt] == pid

        assert on_success.batching
        assert not eventdata_set
        assert not amqpannoated_set

    await consumer.close()
    await receive_thread


@pytest.mark.liveTest
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "flush_after_sending, close_after_sending",
    [(True, False), (False, True), (False, False)],
)
async def test_basic_send_batch_events_round_robin(
    auth_credentials_async, flush_after_sending, close_after_sending, uamqp_transport, client_args
):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    received_events = defaultdict(list)

    async def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )
    receive_thread = asyncio.ensure_future(consumer.receive(on_event=on_event))

    await asyncio.sleep(5)
    sent_events = defaultdict(list)

    async def on_success(events, pid):
        sent_events[pid].extend(events)

    async def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None
    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer:
        partitions = await producer.get_partition_ids()
        partitions_cnt = len(partitions)
        # perform batch sending round-robin
        total_events_cnt = 100
        batch_cnt = partitions_cnt * 2 - 1
        each_partition_cnt = total_events_cnt // batch_cnt
        remain_events = total_events_cnt % batch_cnt
        batches = []
        event_idx = 0
        eventdata_set, amqpannoated_set = set(), set()
        for i in range(batch_cnt):
            batch = await producer.create_batch()
            for j in range(each_partition_cnt // 2):
                event = EventData("test{}:{}".format(i, event_idx))
                event.properties = {"batch_idx": i, "event_idx": event_idx}
                batch.add(event)
                eventdata_set.add(event_idx)
                event_idx += 1
            for j in range(each_partition_cnt // 2, each_partition_cnt):
                event = AmqpAnnotatedMessage(data_body="test{}:{}".format(i, event_idx))
                event.application_properties = {"batch_idx": i, "event_idx": event_idx}
                batch.add(event)
                amqpannoated_set.add(event_idx)
                event_idx += 1
            batches.append(batch)

        # put remain_events in the last batch
        last_batch = await producer.create_batch()
        for i in range(remain_events):
            event = EventData("test:{}:{}".format(len(batches), event_idx))
            event.properties = {"batch_idx": len(batches), "event_idx": event_idx}
            last_batch.add(event)
            eventdata_set.add(event_idx)
            event_idx += 1
        batches.append(last_batch)

        for batch in batches:
            await producer.send_batch(batch)

        if not flush_after_sending and not close_after_sending:
            # ensure it's buffered sending
            for pid in partitions:
                assert len(sent_events[pid]) < each_partition_cnt
            assert sum([len(sent_events[pid]) for pid in partitions]) < total_events_cnt
            # give some time for producer to complete sending and consumer to complete receiving
        else:
            if flush_after_sending:
                await producer.flush()
            if close_after_sending:
                await producer.close()
            # ensure all events are sent
            assert sum([len(sent_events[pid]) for pid in partitions]) == total_events_cnt

        await asyncio.sleep(10)
        assert len(sent_events) == len(received_events) == partitions_cnt

        # ensure all events are received in the correct partition
        for pid in partitions:
            assert len(sent_events[pid]) > 0
            assert len(sent_events[pid]) == len(received_events[pid])
            for i in range(len(sent_events[pid])):
                event = sent_events[pid][i]
                try:  # amqp annotated message
                    event_idx = event.application_properties["event_idx"]
                    amqpannoated_set.remove(event_idx)
                except AttributeError:  # event data
                    event_idx = event.properties["event_idx"]
                    eventdata_set.remove(event_idx)
                assert received_events[pid][i].properties[b"event_idx"] == event_idx

        assert not amqpannoated_set
        assert not eventdata_set
        assert not on_error.err

    await consumer.close()
    await receive_thread


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_with_hybrid_partition_assignment(auth_credentials_async, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    received_events = defaultdict(list)

    async def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )
    receive_thread = asyncio.ensure_future(consumer.receive(on_event=on_event))

    sent_events = defaultdict(list)

    async def on_success(events, pid):
        sent_events[pid].extend(events)

    async def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None
    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer:
        partitions = await producer.get_partition_ids()
        partitions_cnt = len(partitions)
        pid_to_pkey = await random_pkey_generation(partitions)
        expected_event_idx_to_partition = {}
        event_idx = 0
        # 1. send by partition_key, each partition 2 events, two single + one batch containing two
        for pid in partitions:
            pkey = pid_to_pkey[pid]
            await producer.send_event(EventData("{}".format(event_idx)), partition_key=pkey)
            batch = await producer.create_batch(partition_key=pkey)
            batch.add(EventData("{}".format(event_idx + 1)))
            await producer.send_batch(batch)
            for i in range(2):
                expected_event_idx_to_partition[event_idx + i] = pid
            event_idx += 2

        # 2. send by partition_id, each partition 2 events, two single + one batch containing two
        for pid in partitions:
            await producer.send_event(EventData("{}".format(event_idx)), partition_id=pid)
            batch = await producer.create_batch(partition_id=pid)
            batch.add(EventData("{}".format(event_idx + 1)))
            await producer.send_batch(batch)
            for i in range(2):
                expected_event_idx_to_partition[event_idx + i] = pid
            event_idx += 2

        # 3. send without partition, each partition 2 events, two single + one batch containing two
        for _ in partitions:
            await producer.send_event(EventData("{}".format(event_idx)))
            batch = await producer.create_batch()
            batch.add(EventData("{}".format(event_idx + 1)))
            await producer.send_batch(batch)
            event_idx += 2

        await producer.flush()
        assert len(sent_events) == partitions_cnt

        await asyncio.sleep(10)

        visited = set()
        for pid in partitions:
            assert len(sent_events[pid]) == 2 * 3

            for sent_event in sent_events[pid]:
                if int(sent_event.body_as_str()) in expected_event_idx_to_partition:
                    assert expected_event_idx_to_partition[int(sent_event.body_as_str())] == pid

            for recv_event in received_events[pid]:
                if int(sent_event.body_as_str()) in expected_event_idx_to_partition:
                    assert expected_event_idx_to_partition[int(sent_event.body_as_str())] == pid

                assert recv_event.body_as_str() not in visited
                visited.add(recv_event.body_as_str())

        assert len(visited) == 2 * 3 * len(partitions)

    assert not on_error.err
    await consumer.close()
    await receive_thread


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_send_with_timing_configuration(auth_credentials_async, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    received_events = defaultdict(list)

    async def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )
    receive_thread = asyncio.ensure_future(consumer.receive(on_event=on_event))

    sent_events = defaultdict(list)

    async def on_success(events, pid):
        sent_events[pid].extend(events)

    async def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None

    # test max_wait_time
    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        buffered_mode=True,
        max_wait_time=10,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer:
        partitions = await producer.get_partition_ids()
        await producer.send_batch([EventData("data")])
        await asyncio.sleep(5)
        assert not sent_events
        await asyncio.sleep(20)

        assert sum([len(sent_events[pid]) for pid in partitions]) == 1

    assert not on_error.err

    # test max_buffer_length per partition
    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        buffered_mode=True,
        max_wait_time=1000,
        max_buffer_length=10,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    sent_events.clear()
    received_events.clear()
    async with producer:
        partitions = await producer.get_partition_ids()
        for i in range(7):
            await producer.send_event(EventData("data"), partition_id="0")
        assert not sent_events
        batch = await producer.create_batch(partition_id="0")
        for i in range(9):
            batch.add(EventData("9"))
        await producer.send_batch(batch)  # will flush 7 events and put the batch in buffer
        assert sum([len(sent_events[pid]) for pid in partitions]) == 7
        for i in range(5):
            await producer.send_event(
                EventData("data"), partition_id="0"
            )  # will flush batch (9 events) + 1 event, leaving 4 in buffer
        assert sum([len(sent_events[pid]) for pid in partitions]) == 17
        await producer.flush()
        assert sum([len(sent_events[pid]) for pid in partitions]) == 21

    await asyncio.sleep(5)
    assert sum([len(received_events[pid]) for pid in partitions]) == 21
    assert not on_error.err
    await consumer.close()
    await receive_thread


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_long_sleep(auth_credentials_async, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    received_events = defaultdict(list)

    async def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )

    receive_thread = asyncio.ensure_future(consumer.receive(on_event=on_event))

    sent_events = defaultdict(list)

    async def on_success(events, pid):
        sent_events[pid].extend(events)

    async def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None  # ensure no error
    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer:
        await producer.send_event(EventData("test"), partition_id="0")
        await asyncio.sleep(220)
        await producer.send_event(EventData("test"), partition_id="0")
        await asyncio.sleep(5)

    assert not on_error.err
    assert len(sent_events["0"]) == 2
    assert len(received_events["0"]) == 2

    await consumer.close()
    await receive_thread


@pytest.mark.skip("not testing correctly + flaky, fix during MQ")
@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_long_wait_small_buffer(auth_credentials_async, uamqp_transport, client_args):
    fully_qualified_namespace, eventhub_name, credential = auth_credentials_async
    received_events = defaultdict(list)

    async def on_event(partition_context, event):
        received_events[partition_context.partition_id].append(event)

    consumer = EventHubConsumerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        consumer_group="$default",
        uamqp_transport=uamqp_transport,
        **client_args
    )

    receive_thread = asyncio.ensure_future(consumer.receive(on_event=on_event))

    sent_events = defaultdict(list)

    async def on_success(events, pid):
        sent_events[pid].extend(events)

    async def on_error(events, pid, err):
        on_error.err = err

    on_error.err = None  # ensure no error
    producer = EventHubProducerClient(
        fully_qualified_namespace=fully_qualified_namespace,
        eventhub_name=eventhub_name,
        credential=credential(),
        buffered_mode=True,
        on_success=on_success,
        on_error=on_error,
        auth_timeout=3,
        retry_total=3,
        retry_mode="fixed",
        retry_backoff_factor=0.01,
        max_wait_time=10,
        max_buffer_length=100,
        uamqp_transport=uamqp_transport,
        **client_args
    )

    async with producer:
        for i in range(100):
            await producer.send_event(EventData("test"))

    await asyncio.sleep(60)

    assert not on_error.err
    assert sum([len(sent_events[key]) for key in sent_events]) == 100
    assert sum([len(received_events[key]) for key in received_events]) == 100

    await consumer.close()
    await receive_thread
